@@ -7,11 +7,11 @@
 
 /* Includes. ================================================================*/
 
-#include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <setjmp.h>
 #include <string.h>
+#include <glib.h>
 #include "basic.h"
 #include "pools.h"
 #include "values.h"
@@ -30,9 +30,9 @@ typedef struct tree_node /* A rule application is stored in "tree_node". */
   struct tree_node *sibling; /* Alternative tree node. */
   tree_node_type_t type; /* Type of this tree node. */
   int_t rule; /* Number of the executed rule. */
-  int_t index; /* Index of this tree node. */
-  value_t link_fs; /* Feature structure of the link. */
-  value_t result_fs; /* Result feature structure of the resulting state. */
+  int_t state_index; /* Index of this node's state or -1 if node is break. */
+  value_t link_feat; /* Feature structure of the link. */
+  value_t result_feat; /* Result feature structure of the resulting state. */
   int_t rule_set; /* Successor rules of resulting state (-1 for end state). */
   string_t input; /* The input that is not yet analysed. */
 } tree_node_t;
@@ -40,7 +40,7 @@ typedef struct tree_node /* A rule application is stored in "tree_node". */
 typedef struct /* A state in morphological or syntactical analysis. */
 { 
   list_node_t *next;
-  value_t fs; /* Feature structure of input read in so far. */
+  value_t feat; /* Feature structure of input read in so far. */
   string_t input; /* Pointer to input that is analysed next. */
   int_t rule_set; /* Set of rules to be applied. */
   tree_node_t *tree_node; /* Tree node of rule application that created
@@ -68,7 +68,9 @@ bool_t recognised_by_combi_rules;
 bool_t recognised_by_robust_rule; 
 string_t last_analysis_input; 
 void (*debug_state)( int_t index, bool_t enter );
-string_t (*get_surface)( surface_t surface_type );
+char_t * (*get_surface)( surface_t surface_type );
+int_t mor_pruning_min;
+int_t syn_pruning_min;
 
 /* Variables. ===============================================================*/
 
@@ -78,7 +80,6 @@ static analysis_t *analyses[2];
 /* The data structure used to save the analysis tree. */
 static tree_node_t *root_tree_node; /* A pointer to the root tree node. */
 static pool_t tree_pool; /* Pool where tree nodes are stored. */
-static int_t tree_node_count;
 
 static state_t *next_result_state; /* Needed for "next_analysis_result". */
 static tree_node_t *next_tree_node; /* Needed for "get_next_analysis_node". */
@@ -89,9 +90,10 @@ static string_t state_surface, link_surface, link_surface_end;
 static struct /* Information needed to generate states and tree nodes. */
 { 
   analysis_t *analysis;
-  tree_mode_t tree_mode; /* Are we building an analysis tree? */
+  bool_t count_states;
+  bool_t create_tree;
   int_t rule; /* Rule just executed. */
-  value_t link_fs; /* Link's feature structure. */
+  value_t link_feat; /* Link's feature structure. */
   tree_node_t *parent; /* Predecessor tree node. */
   int_t item_index; /* Index of item that is added. */
   string_t input; /* End of analysed input. */
@@ -111,13 +113,6 @@ set_analysis_option( analysis_option_t selected, bool_t setting )
     if (rule_system[ MORPHOLOGY ]->robust_rule == -1) 
       complain( "No morphology \"robust_rule\"." );
     break;
-  case PRUNING_RULE_OPTION:
-    if (rule_system[ SYNTAX ] == NULL 
-	|| rule_system[ SYNTAX ]->pruning_rule == -1)
-    {
-      complain( "No syntax \"pruning_rule\"." );
-    }
-    break;
   case MOR_OUT_FILTER_OPTION:
     if (rule_system[ MORPHOLOGY ]->output_filter == -1) 
       complain( "No morphology \"output_filter\"." );
@@ -138,7 +133,7 @@ set_analysis_option( analysis_option_t selected, bool_t setting )
     break;
   case SYN_INCOMPLETE_OPTION:
     if (rule_system[ SYNTAX ] == NULL)
-      complain ( "No syntax rules loaded." );
+      complain( "No syntax rules loaded." );
     break;
   case CACHE_OPTION:
   case MOR_INCOMPLETE_OPTION:
@@ -161,14 +156,16 @@ get_analysis_option( analysis_option_t selected )
 /* Functions for segmentation and preprocessing. ============================*/
 
 void 
-preprocess_input( string_t input, bool_t expect_quotes )
+preprocess_input( char_t *input, bool_t expect_quotes )
 /* Delete heading and trailing spaces in INPUT
  * and compress all whitespace sequences to a single space.
  * If EXPECT_QUOTES == TRUE, expect quoted input and remove the quotes. */
 { 
-  string_t input_p, output_p;
-  int_t code, i;
-
+  string_t input_p;
+  char_t *output_p;
+  int_t i;
+  u_int_t code;
+  
   output_p = input;
 
   /* Cut heading spaces. */
@@ -188,32 +185,33 @@ preprocess_input( string_t input, bool_t expect_quotes )
 	input_p++;
 	if (*input_p == '\\' || *input_p == '\"')
 	  *output_p++ = *input_p++;
-	else if (TO_LOWER( *input_p ) == 'x') 
+	else if (*input_p >= '0' && *input_p <= '8') 
 	{ 
 	  code = 0;
-	  input_p++;
-	  for (i = 0; i < 2 * (int_t) sizeof( char_t ); i++) 
+	  for (i = 0; i < 3; i++) 
 	  { 
-	    if (IS_DIGIT( *input_p ))
-	      code = 16 * code + ORD( *input_p ) - '0';
-	    else if (TO_LOWER( *input_p ) >= 'a' 
-		     && TO_LOWER( *input_p ) <= 'f')
-	    {
-	      code = 16 * code + 10 + TO_LOWER( *input_p ) - 'a';
-	    }
+	    if (*input_p >= '0' && *input_p <= '9')
+	      code = 8 * code + *input_p++ - '0';
 	    else 
-	      complain( "Control char sequence must have 2 digits." );
-	    input_p++;
+	      complain( "Escape sequence must have 3 octal digits." );
 	  }
-	  *output_p++ = code;
+	  if (! g_unichar_validate( code ))
+	    complain( "Unicode escape sequence defines invalid character." );
+	  output_p += g_unichar_to_utf8( code, output_p );
 	} 
 	else 
 	  complain( "Illegal escape sequence." );
       }
       else /* No escape char. */
-	*output_p++ = *input_p++;
+      {
+	/* We must copy a whole UTF-8 character since next_non_space() below
+	 * expects to be at UTF-8 character boundary. */
+	code = g_utf8_get_char( input_p );
+	input_p = g_utf8_next_char( input_p );
+	output_p += g_unichar_to_utf8( code, output_p );
+      }
     }
-    input_p++;
+    input_p++; /* Read over closing double quotes. */
     input_p = next_non_space( input_p );
     if (*input_p != EOS)
       complain( "Junk after quoted string in input." );
@@ -222,19 +220,23 @@ preprocess_input( string_t input, bool_t expect_quotes )
   {
     while (*input_p != EOS) 
     { 
-      if (IS_SPACE( *input_p )) 
+      code = g_utf8_get_char( input_p );
+      if (g_unichar_isspace( code ))
       { 
 	/* Overread all whitespace and write a single space. */
 	input_p = next_non_space( input_p );
 	*output_p++ = ' ';
       } 
       else 
-	*output_p++ = *input_p++;
+      {
+	input_p = g_utf8_next_char( input_p );
+	output_p += g_unichar_to_utf8( code, output_p );
+      }
     }
   }
 
   /* Last space may be superfluous. */
-  if (output_p > input && IS_SPACE( output_p[-1] )) 
+  if (output_p > input && output_p[-1] == ' ') 
     output_p--;
   *output_p = EOS;
 }
@@ -257,13 +259,13 @@ word_may_end_here( string_t string, rule_t *rule )
 static state_t *
 insert_state( analysis_t *analysis,
 	      list_t *state_list,
-	      value_t fs,
+	      value_t feat,
 	      string_t input,
 	      int_t rule_set,
 	      int_t item_index )
-/* Insert a state, composed of FS, INPUT, RULE_SET, and ITEM_INDEX
- * in the list STATE_LIST, in front of all states with a
- * higher INPUT index. Return this state. */
+/* Insert a state, composed of FEAT, INPUT, RULE_SET, and ITEM_INDEX in the
+ * list STATE_LIST, in front of all states with a higher INPUT index. Return
+ * this state. */
 { 
   state_t *state, *prev_state, *next_state;
 
@@ -272,7 +274,7 @@ insert_state( analysis_t *analysis,
     state = (state_t *) get_pool_space( analysis->state_pool, 1, NULL );
 
   /* Set values. */
-  state->fs = fs;
+  state->feat = feat;
   state->input = input;
   state->rule_set = rule_set;
   state->item_index = item_index;
@@ -293,11 +295,11 @@ insert_state( analysis_t *analysis,
 /*---------------------------------------------------------------------------*/
 
 static tree_node_t *
-add_tree_node( value_t result_fs, 
+add_tree_node( value_t result_feat, 
 	       string_t input, 
 	       int_t rule_set,
 	       tree_node_type_t type )
-/* Add a tree node for a rule that created a state (RESULT_FS and RULE_SET),
+/* Add a tree node for a rule that created a state (RESULT_FEAT and RULE_SET),
  * where INPUT is yet to be analysed. */
 { 
   tree_node_t **tree_node_p;
@@ -310,9 +312,12 @@ add_tree_node( value_t result_fs,
   tree_node->sibling = NULL;
   tree_node->type = type;
   tree_node->rule = state_info.rule;
-  tree_node->index = tree_node_count;
-  tree_node->link_fs = state_info.link_fs;
-  tree_node->result_fs = result_fs;
+  if (type == BREAK_NODE)
+    tree_node->state_index = -1;
+  else
+    tree_node->state_index = state_count;
+  tree_node->link_feat = state_info.link_feat;
+  tree_node->result_feat = result_feat;
   tree_node->rule_set = rule_set;
   tree_node->input = input;
 
@@ -322,55 +327,54 @@ add_tree_node( value_t result_fs,
     tree_node_p = &(*tree_node_p)->sibling;
   *tree_node_p = tree_node;
 
-  /* Increment the number of tree nodes. */
-  tree_node_count++;
-
   return tree_node;
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-add_state( list_t *list, string_t input, value_t fs, int_t rule_set, 
+add_state( list_t *list, string_t input, value_t feat, int_t rule_set, 
 	   tree_node_type_t type )
-/* Add state, consisting of INPUT, FS and RULE_SET, to LIST.
- * When STATE_INFO.TREE_MODE == BUILD_TREE, also generate a tree node. */
+/* Add state, consisting of INPUT, FEAT and RULE_SET, to LIST.
+ * When STATE_INFO.CREATE_TREE == TRUE, also generate a tree node. */
 { 
-  value_t new_fs;
+  value_t new_feat;
   state_t *state;
   
   /* Preserve the feature structure. */
-  new_fs = copy_value_to_pool( state_info.analysis->value_pool, fs, NULL );
+  new_feat = copy_value_to_pool( state_info.analysis->value_pool, feat, NULL );
 
   /* Create a new state. */
-  state = insert_state( state_info.analysis, list, new_fs, input,
+  state = insert_state( state_info.analysis, list, new_feat, input,
                         rule_set, state_info.item_index );
-  if (state_info.tree_mode == BUILD_TREE) 
-    state->tree_node = add_tree_node( new_fs, input, rule_set, type );
+  if (state_info.create_tree) 
+    state->tree_node = add_tree_node( new_feat, input, rule_set, type );
+  if (state_info.count_states)    
+    state_count++;
 }
 
 /* Callback functions needed by "rules.c" ===================================*/
 
 static void 
-add_allo_local( string_t surface, value_t fs )
-/* Add a state, consisting of SURFACE and FS, as an end state. */
+add_allo_local( string_t surface, value_t feat )
+/* Add a state, consisting of SURFACE and FEAT, as an end state. */
 { 
   int_t length;
 
   length = strlen( surface );
   if (length == 0) 
     complain( "Surface is empty." );
-  if (strncmp_no_case( state_surface, surface, length ) != 0) 
+  if (strncmp_no_case( state_surface, surface, length ) != 0)
     complain( "Surface does not match input." );
   add_state( &state_info.analysis->end_states, 
-	     state_surface + length, fs, -1, FINAL_NODE );
+	     state_surface + length, feat, -1, FINAL_NODE );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-add_end_state_local( value_t fs )
-/* Add a state, consisting of FS, as an end state. */
+add_end_state_local( value_t feat )
+/* Add a state, consisting of FEAT, as an end state. */
 { 
   value_t value;
   rule_t *rule;
@@ -382,12 +386,12 @@ add_end_state_local( value_t fs )
       || word_may_end_here( state_info.input, rule ))
   { 
     add_state( &state_info.analysis->end_states,
-	       state_info.input, fs, -1, FINAL_NODE );
+	       state_info.input, feat, -1, FINAL_NODE );
   } 
-  else if (state_info.tree_mode == BUILD_TREE) 
+  else if (state_info.create_tree) 
   { 
     /* Preserve the feature structure. */
-    value = copy_value_to_pool( state_info.analysis->value_pool, fs, NULL );
+    value = copy_value_to_pool( state_info.analysis->value_pool, feat, NULL );
     add_tree_node( value, state_info.input, -1, UNFINAL_NODE );
   }
 }
@@ -395,16 +399,16 @@ add_end_state_local( value_t fs )
 /*---------------------------------------------------------------------------*/
 
 static void 
-add_running_state_local( value_t fs, int_t rule_set )
-/* Add a running state, consisting of FS and RULE_SET. */
+add_running_state_local( value_t feat, int_t rule_set )
+/* Add a running state, consisting of FEAT and RULE_SET. */
 { 
   add_state( &state_info.analysis->running_states, 
-	     state_info.input, fs, rule_set, INTER_NODE );
+	     state_info.input, feat, rule_set, INTER_NODE );
 }
 
 /*---------------------------------------------------------------------------*/
 
-static string_t 
+static char_t *
 get_surface_local( surface_t surface_type )
 /* Return surface SURFACE_TYPE for currently executed rule.
  * The result must be freed after use. */
@@ -420,6 +424,8 @@ get_surface_local( surface_t surface_type )
       state_surf_end = link_surface;
     return new_string_readable( state_surface, state_surf_end );
   case LINK_SURFACE:
+    if (link_surface_end == link_surface)
+      return NULL;
     return new_string_readable( link_surface, link_surface_end );
   case RESULT_SURFACE:
     return new_string_readable( state_surface, link_surface_end );
@@ -538,7 +544,7 @@ next_analysis_result( void )
 
   if (next_result_state == NULL) 
     return NULL;
-  result = next_result_state->fs;
+  result = next_result_state->feat;
   next_result_state = (state_t *) next_result_state->next;
   return result;
 }
@@ -580,14 +586,14 @@ get_next_analysis_node( void )
   if (next_tree_node == NULL) 
     return NULL;
   node = new_mem( sizeof( analysis_node_t ) );
-  node->index = next_tree_node->index;
+  node->index = next_tree_node->state_index;
   node->type = next_tree_node->type;
 
   /* Set parent index. */
   if (next_tree_node->parent == NULL) 
     node->parent_index = -1;
   else 
-    node->parent_index = next_tree_node->parent->index;
+    node->parent_index = next_tree_node->parent->state_index;
 
   /* Set rule name. */
   if (next_tree_node->rule != -1) 
@@ -607,14 +613,14 @@ get_next_analysis_node( void )
     link_surf = next_non_space( next_tree_node->parent->input );
   if (link_surf < next_tree_node->input) 
     node->link_surf = new_string( link_surf, next_tree_node->input );
-  node->link_fs = next_tree_node->link_fs;
+  node->link_feat = next_tree_node->link_feat;
 
   /* Set result surface and feature structure. */
   node->result_surf = new_string( last_analysis_input, next_tree_node->input );
-  node->result_fs = next_tree_node->result_fs;
+  node->result_feat = next_tree_node->result_feat;
 
   /* Set rule set. */
-  if (next_tree_node->result_fs != NULL) 
+  if (next_tree_node->result_feat != NULL) 
     node->rule_set = rule_set_readable( rule_sys, next_tree_node->rule_set );
 
   /* Update NEXT_TREE_NODE. */
@@ -693,7 +699,7 @@ put_into_cache( analysis_t *analysis, string_t input )
 /* Store the results in ANALYSIS for the word form that starts at INPUT
  * in the cache. */
 {
-  value_t *fs_vector;
+  value_t *feat_vector;
   int_t i;
   state_t *state;
   string_t input_end;
@@ -712,15 +718,15 @@ put_into_cache( analysis_t *analysis, string_t input )
   }
 
   /* Allocate a new vector which takes the feature structures. */
-  fs_vector = new_vector( sizeof( value_t ), i );
+  feat_vector = new_vector( sizeof( value_t ), i );
   i = 0;
   FOREACH( state, analysis->end_states ) 
   { 
-    fs_vector[i] = new_value( state->fs );
+    feat_vector[i] = new_value( state->feat );
     i++;
   }
 
-  enter_in_cache( input, input_end, i, fs_vector );
+  enter_in_cache( input, input_end, i, feat_vector );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -744,7 +750,8 @@ execute_robust_rule( analysis_t *analysis,
 
   /* Setup STATE_INFO. */
   state_info.analysis = analysis;
-  state_info.tree_mode = NO_TREE;
+  state_info.count_states = FALSE;
+  state_info.create_tree = FALSE;
   state_info.item_index = 1;
   state_info.input = input_end;
 
@@ -784,7 +791,7 @@ execute_filter_rule( analysis_t *analysis,
     { 
       state = remove_first_node( &old_end_states );
       add_node( &analysis->free_states, (list_node_t *) state, LIST_END );
-      push_value( state->fs );
+      push_value( state->feat );
     }
     build_list( top );
 
@@ -792,7 +799,8 @@ execute_filter_rule( analysis_t *analysis,
 
     /* Execute filter rule. */
     state_info.analysis = analysis;
-    state_info.tree_mode = NO_TREE;
+    state_info.count_states = FALSE;
+    state_info.create_tree = FALSE;
     state_info.item_index = 0;
     state_info.input = input;
     execute_rule( rule_sys, filter_rule );
@@ -814,7 +822,6 @@ execute_pruning_rule( analysis_t *analysis, grammar_t grammar )
 
   state = (state_t *) analysis->running_states.first;
   input = state->input;
-  rule_sys = rule_system[ grammar ];
 
   /* Create a list that contains the results. */
   top = 0;
@@ -824,10 +831,14 @@ execute_pruning_rule( analysis_t *analysis, grammar_t grammar )
     if (state->input != input) 
       break;
     result_count++;
-    push_value( state->fs );
+    push_value( state->feat );
   }
+  /* Don't execute if number of states is too low. */
+  if (result_count < (grammar == SYNTAX ? syn_pruning_min : mor_pruning_min))
+    return;
   build_list( result_count );
 
+  rule_sys = rule_system[ grammar ];
   link_surface = link_surface_end = input; /* Set debugging information. */
   execute_rule( rule_sys, rule_sys->pruning_rule ); /* Execute pruning rule. */
 
@@ -861,24 +872,25 @@ static void
 execute_rules( analysis_t *analysis, 
                rule_sys_t *rule_sys, 
                state_t *state, 
-               value_t link_fs,
+               value_t link_feat,
                string_t link_surf,
                string_t link_surf_end, 
-               tree_mode_t tree_mode,
+	       bool_t count_states,
+               bool_t create_tree,
                rule_type_t rule_type )
-/* Execute the successor rules in RULE_SYS for STATE in ANALYSIS.
+/* Execute the successor rules of RULE_TYPE in RULE_SYS for STATE in ANALYSIS.
  * Consume the segment from LINK_SURF to LINK_SURF_END with feature structure
- * LINK_FS. 
- * Enter a tree node if TREE_MODE == BUILD_TREE. */
+ * LINK_FEAT. */
 { 
   int_t *rule_p;
-  bool_t rules_successful;
+  bool_t rules_successful, rules_executed;
   rule_t *rule;
 
   /* Setup STATE_INFO. */
   state_info.analysis = analysis;
-  state_info.tree_mode = tree_mode;
-  state_info.link_fs = link_fs;
+  state_info.count_states = count_states;
+  state_info.create_tree = create_tree;
+  state_info.link_feat = link_feat;
   state_info.parent = state->tree_node;
   state_info.item_index = state->item_index + 1;
   state_info.input = link_surf_end;
@@ -887,14 +899,14 @@ execute_rules( analysis_t *analysis,
   link_surface = link_surf;
   link_surface_end = link_surf_end;
   if (state->tree_node != NULL) 
-    current_state = state->tree_node->index;
+    current_state = state->tree_node->state_index;
 
   /* Check if we are now executing the rules for a state to be debugged */
   if (debug_state != NULL && state->tree_node != NULL) 
-    debug_state( state->tree_node->index, TRUE );
+    debug_state( state->tree_node->state_index, TRUE );
 
   /* Execute rules in rule set. */
-  rules_successful = FALSE;
+  rules_executed = rules_successful = FALSE;
   for (rule_p = rule_sys->rule_sets + state->rule_set; *rule_p != -1; rule_p++)
   { 
     if (*rule_p == -2) 
@@ -911,10 +923,10 @@ execute_rules( analysis_t *analysis,
       { 
 	state_info.rule = *rule_p;
 	top = 0;
-	push_value( state->fs );
+	push_value( state->feat );
 	if (rule->type == COMBI_RULE) 
 	{ 
-	  push_value( link_fs );
+	  push_value( link_feat );
 	  if (rule->param_count >= 3) 
 	    push_string_value( link_surf, link_surf_end );
 	  if (rule->param_count >= 4) 
@@ -926,6 +938,7 @@ execute_rules( analysis_t *analysis,
 	    push_string_value( link_surf, NULL );
 	}
 	execute_rule( rule_sys, *rule_p );
+	rules_executed = TRUE;
 	rules_successful |= rule_successful;
       }
     }
@@ -934,10 +947,10 @@ execute_rules( analysis_t *analysis,
   
   /* Save the current debug mode if we are leaving a debug state. */
   if (debug_state != NULL && state->tree_node != NULL) 
-    debug_state( state->tree_node->index, FALSE );
+    debug_state( state->tree_node->state_index, FALSE );
 
-  /* Enter a tree node for a rule set that did not fire. */
-  if (! rules_successful && tree_mode == BUILD_TREE && rule_type != END_RULE) 
+  /* Enter a tree node if rules where executed but did not fire. */
+  if (rules_executed && ! rules_successful && create_tree) 
   { 
     state_info.rule = -1;
     add_tree_node( NULL, link_surf_end, -1, BREAK_NODE );
@@ -949,13 +962,13 @@ execute_rules( analysis_t *analysis,
 static void
 check_end_states( analysis_t *analysis, 
 		  grammar_t grammar, 
-		  analysis_mode_t mode )
-/* If MODE == ANALYSE_ALL, 
+		  bool_t analyse_all )
+/* If ANALYSE_ALL == TRUE,
  * delete all states in ANALYSIS that didn't consume all the input. */
 { 
   state_t *state;
 
-  if (mode != ANALYSE_ALL
+  if (! analyse_all
       || (grammar == MORPHOLOGY && options[ MOR_INCOMPLETE_OPTION ])
       || (grammar == SYNTAX && options[ SYN_INCOMPLETE_OPTION ]))
   { 
@@ -978,28 +991,28 @@ check_end_states( analysis_t *analysis,
 void 
 analyse( grammar_t grammar, 
          string_t input, 
-         tree_mode_t tree_mode,
-         analysis_mode_t mode )
+         bool_t create_tree,
+         bool_t analyse_all )
 /* Perform a LAG analysis of INPUT using GRAMMAR (MORPHOLOGY or SYNTAX).
- * An analysis tree will be built if TREE_MODE == BUILD_TREE.
- * The whole input will be analysed if MODE == ANALYSE_ALL. */
+ * An analysis tree will be built if CREATE_TREE == TRUE.
+ * The whole input will be analysed if ANALYSE_ALL == TRUE. */
 { 
   rule_sys_t *rule_sys;
   state_t *initial_state;
   analysis_t *analysis;
   state_t *state;
   string_t current_input;
-  value_t link_fs;
+  value_t link_feat;
   string_t link_surf_end; /* End of the link's surface. */
   string_t input_behind_space;
   state_t *mor_state; /* Morphology end state. */
   bool_t use_cache;
 
-  if (mode == ANALYSE_ALL) 
+  if (analyse_all) 
   { 
     top_grammar = grammar;
     root_tree_node = NULL;
-    state_count = 0;
+    state_count = 1; /* We will insert the initial state. */
     last_analysis_input = input;
     recognised_by_robust_rule = recognised_by_combi_rules = FALSE;
   }
@@ -1028,16 +1041,16 @@ analyse( grammar_t grammar,
   /* Try to get analysis result from cache. */
   use_cache = (grammar == MORPHOLOGY 
 	       && options[ CACHE_OPTION ] 
-	       && tree_mode == NO_TREE 
-	       && (mode == ANALYSE_WORD || *get_word_end( input ) == EOS));
+	       && ! create_tree
+	       && (! analyse_all || *get_word_end( input ) == EOS));
   if (use_cache && get_from_cache( analysis, input )) 
     return;
 
   /* Enter the initial state. */
   initial_state = insert_state( analysis, &analysis->running_states,
-                                rule_sys->values + rule_sys->initial_fs,
+                                rule_sys->values + rule_sys->initial_feat,
                                 input, rule_sys->initial_rule_set, 0 );
-  if (tree_mode == BUILD_TREE) 
+  if (create_tree) 
   { 
     /* Clear all tree nodes and setup ROOT_TREE_NODE. */
     clear_pool( tree_pool );
@@ -1047,13 +1060,12 @@ analyse( grammar_t grammar,
     root_tree_node->sibling = NULL;
     root_tree_node->type = INTER_NODE;
     root_tree_node->rule = -1;
-    root_tree_node->index = 0;
-    root_tree_node->link_fs = NULL;
-    root_tree_node->result_fs = rule_sys->values + rule_sys->initial_fs;
+    root_tree_node->state_index = 0;
+    root_tree_node->link_feat = NULL;
+    root_tree_node->result_feat = rule_sys->values + rule_sys->initial_feat;
     root_tree_node->rule_set = rule_sys->initial_rule_set;
     root_tree_node->input = input;
     initial_state->tree_node = root_tree_node;
-    tree_node_count = 1;
   }
 
   /* Analyse while there are running states. */
@@ -1061,9 +1073,10 @@ analyse( grammar_t grammar,
   { 
     state = (state_t *) analysis->running_states.first;
     current_input = state->input;
-    if (options[ PRUNING_RULE_OPTION ] && current_input > input 
-        && rule_sys->pruning_rule != -1) 
-    { 
+    if (((grammar == MORPHOLOGY && mor_pruning_min > 0)
+	 || (grammar == SYNTAX && syn_pruning_min > 0))
+	&& current_input > input && rule_sys->pruning_rule != -1) 
+    {
       execute_pruning_rule( analysis, grammar ); 
     }
     if (current_input > input) /* Apply end_rules if any input was parsed. */
@@ -1074,7 +1087,7 @@ analyse( grammar_t grammar,
 	if (state->input != current_input) 
 	  break;
         execute_rules( analysis, rule_sys, state, NULL, current_input, 
-                       current_input, tree_mode, END_RULE );
+                       current_input, analyse_all, create_tree, END_RULE );
       }
     }
     if (*current_input == EOS) 
@@ -1084,15 +1097,15 @@ analyse( grammar_t grammar,
       /* Look for prefixes of increasing length
        * that match the string at CURRENT_INPUT. */
       search_for_prefix( current_input );
-      while (get_next_prefix( &link_surf_end, &link_fs )) 
+      while (get_next_prefix( &link_surf_end, &link_feat )) 
       { 
 	/* Combine that link with all morphological states. */
         FOREACH( state, analysis->running_states ) 
         { 
 	  if (state->input != current_input) 
 	    break;
-          execute_rules( analysis, rule_sys, state, link_fs, current_input,
-                         link_surf_end, tree_mode, COMBI_RULE );
+          execute_rules( analysis, rule_sys, state, link_feat, current_input,
+                         link_surf_end, analyse_all, create_tree, COMBI_RULE );
         }
       }
     } 
@@ -1101,7 +1114,7 @@ analyse( grammar_t grammar,
       input_behind_space = next_non_space( current_input );
 
       /* Call morphological analysis to get feature structures for the link. */
-      analyse( MORPHOLOGY, input_behind_space, NO_TREE, ANALYSE_WORD );
+      analyse( MORPHOLOGY, input_behind_space, FALSE, FALSE );
 
       /* Execution of morphology rules has changed STATE_SURFACE. */
       state_surface = input;
@@ -1110,8 +1123,8 @@ analyse( grammar_t grammar,
       FOREACH( mor_state, analyses[ MORPHOLOGY ]->end_states ) 
       { 
 	/* The morphology will be cleared by next morphological analysis,
-         * so copy STATE->FS to the syntax pool. */ 
-        link_fs = copy_value_to_pool( analysis->value_pool, mor_state->fs, 
+         * so copy STATE->FEAT to the syntax pool. */ 
+        link_feat = copy_value_to_pool( analysis->value_pool, mor_state->feat, 
 				      NULL );
 
         /* Combine the link with all syntactic states. */
@@ -1119,9 +1132,9 @@ analyse( grammar_t grammar,
 	{ 
 	  if (state->input != current_input) 
 	    break;
-  	  execute_rules( analysis, rule_sys, state, link_fs,
+  	  execute_rules( analysis, rule_sys, state, link_feat, 
 			 input_behind_space, mor_state->input, 
-			 tree_mode, COMBI_RULE );
+			 analyse_all, create_tree, COMBI_RULE );
         }
       }
     }
@@ -1134,22 +1147,20 @@ analyse( grammar_t grammar,
       if (state == NULL || state->input != current_input) 
 	break;
       remove_first_node( &analysis->running_states );
-      if (mode == ANALYSE_ALL) 
-	state_count++;
       add_node( &analysis->free_states, (list_node_t *) state, LIST_END );
     }
   } /* End of loop that consumes all running states. */
 
-  check_end_states( analysis, grammar, mode );
-  if (mode == ANALYSE_ALL && analysis->end_states.first != NULL) 
+  check_end_states( analysis, grammar, analyse_all );
+  if (analyse_all && analysis->end_states.first != NULL) 
     recognised_by_combi_rules = TRUE;
   if (grammar == MORPHOLOGY) 
   { 
     if (analysis->end_states.first == NULL && options[ ROBUST_RULE_OPTION ]) 
     { 
       execute_robust_rule( analysis, rule_sys, input );
-      check_end_states( analysis, grammar, mode );
-      if (mode == ANALYSE_ALL && analysis->end_states.first != NULL) 
+      check_end_states( analysis, grammar, analyse_all );
+      if (analyse_all && analysis->end_states.first != NULL) 
 	recognised_by_robust_rule = TRUE;
     }
     if (options[ MOR_OUT_FILTER_OPTION ]) 
