@@ -16,7 +16,7 @@
 
 /* Global variables. ========================================================*/
 
-char_t malaga_version[] = "7.11";
+char_t malaga_version[] = "7.12";
 string_t program_name;
 bool_t in_emacs_malaga_mode;
 bool_t split_hangul_syllables;
@@ -780,7 +780,7 @@ init_basic( string_t prog_name )
   error_text = new_text();
   program_name = prog_name;
   in_emacs_malaga_mode = (getenv( "MALAGA_MODE" ) != NULL);
-#ifdef WINDOWS
+#ifdef WIN32
   /* We must switch off buffering by hand if we are connected to Emacs. */
   if (getenv( "EMACS" ) != NULL) 
   {
