@@ -18,6 +18,7 @@
 #include <pango/pango.h>
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
+#include <cairo.h>
 #include "basic.h"
 #include "scanner.h"
 #include "input.h"
@@ -492,6 +493,7 @@ struct canvas
   int_t space_width; /* Width of a space in FONT. */
   int_t comma_width; /* Width of a comma in FONT. */
   int_t border_width; /* Width of record border. */
+  int_t line_width; /* Width of canvas lines. */
   int_t border_height; /* Height of list and record borders. */
   bool_t hanging_style; /* TRUE if values are "hanging down" from baseline. */
   bool_t alternate_cursor; /* TRUE if alternate cursor is displayed. */
@@ -510,8 +512,10 @@ bool_t hanging_style;
  * always only one canvas being drawn. */
 static char_t ps_string[ BUFFER_SIZE ]; /* Converted PostScript string. */
 static GdkDrawable *drawable; /* Drawable we currently draw into. */
+static cairo_t *cairo; /* Cairo context for DRAWABLE. */
 static rectangle_t area; /* The clipping rectangle of DRAWABLE, using the
 			  * coordinates of the whole canvas. */
+static double canvas_shift; /* Shift if stroke width is odd. */
 static GdkRectangle clip; /* The clipping rectangle of DRAWABLE, using the
 			   * coordinates of DRAWABLE. */
 static GdkGC *gc; /* Graphics context used to draw. */
@@ -694,6 +698,14 @@ parse_value( void )
 
 /* Displaying Malaga values. ================================================*/
 
+int_t
+get_line_width( canvas_t *canvas )
+/* Get line width in pixels for CANVAS. */
+{
+  return canvas->line_width;
+}
+
+/*---------------------------------------------------------------------------*/
 int_t
 get_space_width( canvas_t *canvas )
 /* Get space width in pixels for CANVAS. */
@@ -742,18 +754,21 @@ set_color( color_t color )
       fprintf( ps_stream, "0 G\n" );
   }
   else 
+  {
     gdk_gc_set_foreground( gc, &colors[ color ] );
+    gdk_cairo_set_source_color( cairo, &colors[ color ] );
+  }
 }
 
 /*---------------------------------------------------------------------------*/
 
 static bool_t
-clip_line( int_t *x1, int_t *y1, int_t *x2, int_t *y2, 
-	   int_t x_min, int_t x_max )
+clip_line( double *x1, double *y1, double *x2, double *y2, 
+	   double x_min, double x_max )
 /* Clip the x-coordinates of the line from (X1,Y1) to (X2,Y2) to the range
  * [X_MIN..X_MAX].  The start point and the end point may be swapped. */ 
 {
-  int_t t;
+  double t;
 
   /* Make sure that *X1 <= *X2. */
   if (*x2 < *x1)
@@ -768,14 +783,14 @@ clip_line( int_t *x1, int_t *y1, int_t *x2, int_t *y2,
   if (*x1 < x_min)
   {
     /* We can be sure that *X1 != *X2, since *X1 < X_MIN, but *X2 >= X_MIN. */
-    *y1 += ((x_min - *x1) * (long long) (*y2 - *y1)) / (*x2 - *x1);
+    *y1 += ((x_min - *x1) * (*y2 - *y1)) / (*x2 - *x1);
     *x1 = x_min;
   }
 
   if (*x2 > x_max)
   {
     /* We can be sure that *X1 != *X2, since *X2 > X_MAX, but *X1 <= X_MAX. */
-    *y2 += ((x_max - *x2) * (long long) (*y2 - *y1)) / (*x2 - *x1);
+    *y2 += ((x_max - *x2) * (*y2 - *y1)) / (*x2 - *x1);
     *x2 = x_max;
   }
 
@@ -796,7 +811,12 @@ draw_lines_clipped( GdkPoint p[], int count )
  * part of the first line, the point (x + dx, y + dy) is part of the second
  * line, and vice versa. */
 {
-  int_t i, x1, y1, x2, y2;
+  int_t i;
+  double x1, y1, x2, y2;
+  const double left = clip.x - 10;
+  const double right = clip.x + clip.width + 10;
+  const double top = clip.y - 10;
+  const double bottom = clip.y + clip.height + 10;
 
   for (i = 1; i < count; i++)
   {
@@ -804,10 +824,11 @@ draw_lines_clipped( GdkPoint p[], int count )
     y1 = p[i-1].y;
     x2 = p[i].x;
     y2 = p[i].y;
-    if (clip_line( &x1, &y1, &x2, &y2, clip.x, clip.x + clip.width - 1 )
-	&& clip_line( &y1, &x1, &y2, &x2, clip.y, clip.y + clip.height - 1 ))
+    if (clip_line( &x1, &y1, &x2, &y2, left, right)
+	&& clip_line( &y1, &x1, &y2, &x2, top, bottom))
     {
-      gdk_draw_line( drawable, gc, x1, y1, x2, y2 );
+      cairo_move_to( cairo, x1 + canvas_shift, y1 + canvas_shift);
+      cairo_line_to( cairo, x2 + canvas_shift , y2 + canvas_shift);
     }
   }
 }
@@ -846,7 +867,7 @@ draw_lines( int_t count, int_t x, int_t y, ... )
   {
     if (count > max_points) 
       max_points = renew_vector( &p, sizeof( GdkPoint ), count );
-    left = right = over = under = need_clipping = FALSE;
+    left = right = over = under = FALSE;
     va_start( args, y );
     for (i = 0; ; i++)
     {
@@ -874,10 +895,15 @@ draw_lines( int_t count, int_t x, int_t y, ... )
     va_end( args );
     if (left && right && over && under) 
     {
+      cairo_new_path( cairo );
       if (need_clipping)
 	draw_lines_clipped( p, count );
       else
-	gdk_draw_lines( drawable, gc, p, count );
+      {
+	for (i = 0; i < count; i++)
+	  cairo_line_to(cairo, p[i].x + canvas_shift, p[i].y + canvas_shift);
+      }
+      cairo_stroke( cairo );
     }
   }
 }
@@ -903,8 +929,12 @@ draw_rectangle( int_t x, int_t y, int_t width, int_t height )
     y -= area.y;
     if (x + width > 0 && x < area.width && y + height > 0 && y < area.height)
     {
-      gdk_draw_rectangle( drawable, gc, TRUE, 
-			  clip.x + x, clip.y + y, width, height );
+      cairo_new_path( cairo );
+      cairo_rectangle( cairo, 
+		       clip.x + x + canvas_shift, 
+		       clip.y + y + canvas_shift, 
+		       width, height );
+      cairo_fill( cairo );
     }
   }
 }
@@ -928,8 +958,14 @@ draw_circle( bool_t filled, int_t x, int_t y, int_t r )
     if (x + r >= 0 && x - r <= area.width 
 	&& y + r >= 0 && y - r <= area.height)
     {
-      gdk_draw_arc( drawable, gc, filled, clip.x + x - r, clip.y + y - r, 
-		    2 * r, 2 * r, 0, 64 * 360 );
+      cairo_new_path( cairo );
+      cairo_arc( cairo, clip.x + x + canvas_shift, clip.y + y + canvas_shift,
+		 r, 0, 2 * M_PI );
+      cairo_close_path( cairo );
+      if (filled) 
+	cairo_fill( cairo ); 
+      else 
+	cairo_stroke( cairo );
     }
   }
 }
@@ -1143,7 +1179,7 @@ draw_layout( canvas_t *canvas, PangoLayout *layout, int_t x, int_t y )
       return;
     if (y + height <= 0 || y >= area.height) 
       return;
-
+    
     gdk_draw_layout( drawable, gc, clip.x + x, clip.y + y, layout );
   }
 }
@@ -1372,7 +1408,13 @@ expose_event( GtkWidget *widget, GdkEventExpose *event, canvas_t *canvas )
   clip.width = area.width = event->area.width;
   clip.height = area.height = event->area.height;
   drawable = canvas->draw_area->window;
+  cairo = gdk_cairo_create(drawable);
+  cairo_set_line_width( cairo, canvas->line_width );
+  cairo_set_line_cap( cairo, CAIRO_LINE_CAP_SQUARE );
+  cairo_set_line_join( cairo, CAIRO_LINE_JOIN_BEVEL );
   canvas->expose( canvas, &area );
+  cairo_destroy(cairo);
+  cairo = NULL;
 
   return TRUE;
 }
@@ -1405,6 +1447,7 @@ configure_draw_area( canvas_t *canvas )
     canvas->comma_width = ps_string_width( "," );
     canvas->font_ascent = 917;
     canvas->font_height = 1150;
+    canvas->line_width = 25;
   }
   else
   {
@@ -1412,8 +1455,10 @@ configure_draw_area( canvas_t *canvas )
 		   &canvas->font_height, &canvas->font_ascent );
     config_layout( canvas, canvas->comma_layout, &canvas->comma_width, 
 		   NULL, NULL );
-    canvas->border_width = 6;
-    canvas->border_height = 3;
+    canvas->line_width = (canvas->font_height + 8) / 12;
+    canvas->border_width = 6 * canvas->line_width;
+    canvas->border_height = 3 * canvas->line_width;
+    canvas_shift = ((canvas->line_width & 1) != 0) ? 0.5 : 0.0;
   }
   canvas->configure( canvas, &canvas->width, &canvas->height );
 }
@@ -1528,7 +1573,7 @@ write_postscript( canvas_t *canvas )
 	   "/S /stroke load def\n"
 	   "/T { /Helvetica-ISOLatin1 1000 selectfont show } bind def\n" );
 
-  fprintf( ps_stream, "25 setlinewidth\n" );
+  fprintf( ps_stream, "%d setlinewidth\n", canvas->line_width );
   fprintf( ps_stream, "%f %f translate\n", PAPER_BORDER, PAPER_BORDER );
   fprintf( ps_stream, "%f %f scale\n", scaling, scaling );
 
