@@ -106,6 +106,9 @@ extern void popup_menu( canvas_t *canvas );
 
 /* Primitives to lay out and draw a canvas. ---------------------------------*/
 
+extern int_t get_line_width( canvas_t *canvas );
+/* Get width of a line in CANVAS. */
+
 extern int_t get_space_width( canvas_t *canvas );
 /* Get width of space character in CANVAS. */
 
