@@ -19,7 +19,7 @@ extern void stop_display_process( void );
 
 /* Commands. ================================================================*/
 
-extern command_t display_line_option;
+extern command_t display_cmd_option;
 /* Set the command line to start the display process. */
 
 /* End of file. =============================================================*/
