@@ -15,13 +15,13 @@
 #include <glib.h>
 #include "basic.h"
 #include "files.h"
-#ifdef UNIX
+#ifdef POSIX
 #include <unistd.h>
 #include <pwd.h>
 #include <sys/mman.h>
 #include <fcntl.h>
 #endif
-#ifdef WINDOWS
+#ifdef WIN32
 #include <windows.h>
 #endif
 
@@ -136,7 +136,7 @@ map_file( string_t file_name, void **address, int_t *length )
  * memory region starting at *ADDRESS and will occupy LENGTH bytes.
  * After usage, return the memory region via "unmap_file". */
 { 
-#ifdef UNIX
+#ifdef POSIX
   int file_descriptor;
 
   /* Get a file descriptor. */
@@ -160,7 +160,7 @@ map_file( string_t file_name, void **address, int_t *length )
   close( file_descriptor );
 #endif
 
-#ifdef WINDOWS
+#ifdef WIN32
   HANDLE file_handle, map_handle;
 
   file_handle = CreateFile( file_name, GENERIC_READ, FILE_SHARE_READ, NULL, 
@@ -190,10 +190,10 @@ unmap_file( void **address, int_t length )
 /* Return the memory region that has been allocated by "map_file".
  * The region starts at *ADDRESS and occupies LENGTH bytes. */
 { 
-#ifdef UNIX
+#ifdef POSIX
   munmap( *address, length );
 #endif
-#ifdef WINDOWS
+#ifdef WIN32
   UnmapViewOfFile( *address );
 #endif
   *address = NULL;
@@ -212,7 +212,7 @@ name_in_path( string_t path_name )
   {
     if (name[-1] == '/') 
       break;
-#ifdef WINDOWS
+#ifdef WIN32
     if (name[-1] == '\\'
         || (name == path_name + 2 && IS_LETTER( name[-2] ) && name[-1] == ':'))
     {
@@ -248,7 +248,7 @@ tidy_path( char_t *path )
   char_t *src_p;
   char_t *dest_p;
 
-#ifdef UNIX
+#ifdef POSIX
   dest_p = src_p = path;
   while (*src_p != EOS) 
   { 
@@ -281,7 +281,7 @@ tidy_path( char_t *path )
   *dest_p = EOS;
 #endif
 
-#ifdef WINDOWS
+#ifdef WIN32
   /* The first two chars is the drive specification. */
   if (! IS_LETTER( path[0] ) || path[1] != ':') 
     complain( "Missing drive name." );
@@ -331,7 +331,7 @@ absolute_path( string_t src_path, string_t relative_to )
  * RELATIVE_TO must be an absolute path name (a directory or a file).
  * The returned path must be freed after use. */
 { 
-#ifdef UNIX
+#ifdef POSIX
   text_t *path;
   string_t src_path_p, login, login_p;
   char_t *dest_path;
@@ -394,7 +394,7 @@ absolute_path( string_t src_path, string_t relative_to )
   return dest_path;
 #endif
 
-#ifdef WINDOWS
+#ifdef WIN32
   text_t *path;
   string_t src_path_p, dest_path;		
   string_t relative_end, relative_dir;
@@ -507,7 +507,7 @@ static string_t extension_start( string_t name )
   {
     if (*t == '/') 
       s = NULL;
-#ifdef WINDOWS
+#ifdef WIN32
     else if (*t == '\\') 
       s = NULL;
 #endif
