@@ -13,6 +13,7 @@
 #include <stdlib.h>
 #include <time.h>
 #include <setjmp.h>
+#include <glib.h>
 #include "basic.h"
 #include "pools.h"
 #include "values.h"
@@ -33,7 +34,7 @@
 /* Macros. ==================================================================*/
 
 #define MARK_LAST_ENTRY(var) ((var) = - ((var) + 1))
-/* Mark last entry of a list in FS_LISTS. */
+/* Mark last entry of a list in FEAT_LISTS. */
 
 /* types ====================================================================*/
 
@@ -41,12 +42,12 @@ typedef struct /* A feature structure for a given lexicon entry. */
 { 
   list_node_t *next; /* Next feature structure. */
   value_t value; /* Feature structure of this entry as a value. */
-} fs_node_t;
+} feat_node_t;
 
 typedef struct /* An entry in the lexicon tree. */
 { 
   avln_node_t node; /* The lexicon tree is an AVLN tree. */
-  list_t fs_list; /* The list of feature structures for this lexicon entry. */
+  list_t feat_list; /* The list of feature structures in this lexicon entry. */
 } lex_node_t;
 
 typedef struct /* A node in the constants tree. */
@@ -64,13 +65,14 @@ string_t lex_entry_file_name; /* Name of lexicon file just parsed. */
 
 rule_sys_t *allo_rule_sys;
 
-static const_node_t *const_tree; /* Root of the tree of constants. */
-
-static lex_node_t *lex_tree; /* Root of the lexicon tree. */
+static avln_node_t *const_tree; /* Root of the tree of constants.
+				 * Actually points to a const_node_t. */
+static avln_node_t *lex_tree; /* Root of the lexicon tree.
+			       * Actually points to a lex_node_t. */
 static pool_t string_pool;
 static pool_t lex_node_pool;
-static pool_t fs_node_pool;
-static list_t fs_free_list;
+static pool_t feat_node_pool;
+static list_t feat_free_list;
 
 static int_t prelex_count; /* Number of prelex entries. */
 static int_t lex_entry_count; /* Number of lexicon entries. */
@@ -87,43 +89,44 @@ static void parse_value_local( void );
 /* Functions for rule execution. ============================================*/
 
 static void 
-free_fs_list( lex_node_t *entry )
+free_feat_list( lex_node_t *entry )
 { 
-  fs_node_t *fs;
+  feat_node_t *feat;
   
   /* Free the feature structures of this lexicon entry. */
-  while (entry->fs_list.first != NULL) 
+  while (entry->feat_list.first != NULL) 
   { 
-    fs = remove_first_node( &entry->fs_list );
-    free_mem( &fs->value );
-    add_node( &fs_free_list, (list_node_t *) fs, LIST_END );
+    feat = remove_first_node( &entry->feat_list );
+    free_mem( &feat->value );
+    add_node( &feat_free_list, (list_node_t *) feat, LIST_END );
   }
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void 
-free_lex_tree_local( lex_node_t *lex_node )
+static void
+free_lex_tree_local( avln_node_t *node )
+/* Free NODE and its descendants.  Do nothing if NODE == NULL. */
 { 
-  if (lex_node == NULL) 
+  if (node == NULL) 
     return;
-  free_lex_tree_local( (lex_node_t *) lex_node->node.left );
-  free_lex_tree_local( (lex_node_t *) lex_node->node.right );
-  free_fs_list( lex_node );
+  free_lex_tree_local( node->left );
+  free_lex_tree_local( node->right );
+  free_feat_list( (lex_node_t *) node );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
 free_lex_tree( void )
-/* Free all memory used by lexicon buffer. */
+/* Free all memory used by the lexicon buffer. */
 { 
   free_lex_tree_local( lex_tree );
   lex_tree = NULL;
-  clear_list( &fs_free_list );
+  clear_list( &feat_free_list );
   clear_pool( lex_node_pool );
   clear_pool( string_pool );
-  clear_pool( fs_node_pool ) ;
+  clear_pool( feat_node_pool ) ;
   lex_entry_count = allomorph_count = 0;
   prelex_count = intermediate_count = -1;
 }
@@ -131,23 +134,24 @@ free_lex_tree( void )
 /*---------------------------------------------------------------------------*/
 
 static void 
-add_to_fs_list( lex_node_t *entry, value_t value )
+add_to_feat_list( lex_node_t *entry, value_t value )
+/* Add the feature structure VALUE to the feature structures in ENTRY. */
 { 
-  fs_node_t *fs;
+  feat_node_t *feat;
 
-  if (fs_free_list.first != NULL) 
-    fs = remove_first_node( &fs_free_list );
+  if (feat_free_list.first != NULL) 
+    feat = remove_first_node( &feat_free_list );
   else 
-    fs = get_pool_space( fs_node_pool, 1, NULL );
-  fs->value = new_value( value );
-  add_node( &entry->fs_list, (list_node_t *) fs, LIST_END );
+    feat = get_pool_space( feat_node_pool, 1, NULL );
+  feat->value = new_value( value );
+  add_node( &entry->feat_list, (list_node_t *) feat, LIST_END );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-lex_add_allo( string_t surface, value_t fs )
-/* Add an allomorph, consisting of SURF and FS, to the allomorph lexicon. */
+lex_add_allo( string_t surface, value_t feat )
+/* Add an allomorph, consisting of SURF and FEAT, to the allomorph lexicon. */
 { 
   lex_node_t *entry;
 
@@ -162,12 +166,12 @@ lex_add_allo( string_t surface, value_t fs )
   { 
     entry = (lex_node_t *) get_pool_space( lex_node_pool, 1, NULL );
     entry->node.name = copy_string_to_pool( string_pool, surface, NULL );
-    clear_list( &entry->fs_list );
+    clear_list( &entry->feat_list );
     insert_avln_node( (avln_node_t *) entry, (avln_node_t **) &lex_tree );
   }
 
   /* Add new feature structure. */
-  add_to_fs_list( entry, fs );
+  add_to_feat_list( entry, feat );
   allomorph_count++;
 }
 
@@ -184,46 +188,47 @@ execute_allo_rule( void )
 /*---------------------------------------------------------------------------*/
 
 static void 
-lex_add_end_state( value_t fs )
-/* Add a filtered allomorph, with feature structure FS, to the allomorph
+lex_add_end_state( value_t feat )
+/* Add a filtered allomorph, with feature structure FEAT, to the allomorph
  * lexicon. */
 { 
-  add_to_fs_list( current_lexicon_entry, fs );
+  add_to_feat_list( current_lexicon_entry, feat );
   allomorph_count++;
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-execute_output_filter_local( lex_node_t *entry )
-/* Execute the output filter on the lexicon tree with root node ENTRY. */
+execute_output_filter_local( avln_node_t *node )
+/* Execute the output filter on the lexicon tree with root NODE. */
 { 
-  fs_node_t *fs;
+  feat_node_t *feat;
+  lex_node_t *lex_node = (lex_node_t *) node;
 
-  if (entry == NULL) 
+  if (node == NULL) 
     return;
 
-  /* Execute the filter for entries alphabetically before ENTRY. */
-  execute_output_filter_local( (lex_node_t *) entry->node.left );
+  /* Execute the filter for entries alphabetically before LEX_NODE. */
+  execute_output_filter_local( node->left );
 
   /* Create a list containing all allomorphs with SURFACE. */
   top = 0;
-  FOREACH( fs, entry->fs_list ) 
-    push_value( fs->value );
+  FOREACH( feat, lex_node->feat_list ) 
+    push_value( feat->value );
   build_list( top );
 
-  /* Execute the output filter rule for ENTRY. */
-  free_fs_list( entry );
-  current_lexicon_entry = entry;
+  /* Execute the output filter rule for LEX_NODE. */
+  free_feat_list( lex_node );
+  current_lexicon_entry = lex_node;
   execute_rule( allo_rule_sys, allo_rule_sys->output_filter );
   if (! rule_successful) 
   { 
     complain( "Output filter generated no allomorphs for \"%s\".", 
-	      entry->node.name );
+	      node->name );
   }
 
-  /* Execute the filter for entries alphabetically behind ENTRY. */
-  execute_output_filter_local( (lex_node_t *) entry->node.right );
+  /* Execute the filter for entries alphabetically behind LEX_NODE. */
+  execute_output_filter_local( node->right );
 }
 
 /*---------------------------------------------------------------------------*/
@@ -233,7 +238,7 @@ execute_output_filter( void )
 /* Execute the lexicon output filter on all entries in the lexicon tree. */
 { 
   /* If there's no allomorph filter rule, we're finished. */
-  if (allo_rule_sys->output_filter == -1) 
+  if (allo_rule_sys->output_filter == -1)
     return;
 
   add_end_state = lex_add_end_state;
@@ -246,14 +251,17 @@ execute_output_filter( void )
 /* Support functions for parsing. ===========================================*/
 
 static void 
-free_const_node( const_node_t **const_node_p )
+free_const_node( avln_node_t **const_node_p )
+/* Free *CONST_NODE_P and its descendants, and set *CONST_NODE_P to NULL. */
 { 
+  const_node_t *const_node = (const_node_t *) *const_node_p;
+
   if (*const_node_p == NULL) 
     return;
-  free_const_node( (const_node_t **) &(*const_node_p)->node.left );
-  free_const_node( (const_node_t **) &(*const_node_p)->node.right );
-  free_mem( &(*const_node_p)->node.name );
-  free_mem( &(*const_node_p)->value );
+  free_const_node( &(*const_node_p)->left );
+  free_const_node( &(*const_node_p)->right );
+  free_mem( &(*const_node_p)->name );
+  free_mem( &const_node->value );
   free_mem( const_node_p );
 }
 
@@ -538,34 +546,35 @@ parse_lex_values( void )
 /* Functions for construction of run-time lexicon. ==========================*/
 
 static void 
-print_lex_tree_local( FILE *stream, string_t allo_format, lex_node_t *entry,
-                      int_t *count )
-/* Print all lexicon entries in tree ENTRY to STREAM using allomorph format
+print_lex_tree( FILE *stream, string_t allo_format, avln_node_t *node,
+		int_t *count )
+/* Print all lexicon entries in tree NODE to STREAM using allomorph format
  * ALLO_FORMAT. *COUNT is the counter of entries that have been printed.
  * If ALLO_FORMAT == NULL, use line breaking. */
 { 
-  fs_node_t *fs;
-  string_t line_number, surface, buffer, value_string;
+  feat_node_t *feat;
+  string_t line_number, buffer, value_string;
+  char_t *surface;
+  lex_node_t *lex_node = (lex_node_t *) node;
 
-  if (entry == NULL) 
+  if (node == NULL) 
     return;
-  print_lex_tree_local( stream, allo_format, (lex_node_t *) entry->node.left, 
-			count );
-  surface = new_string_readable( entry->node.name, NULL );
+  print_lex_tree( stream, allo_format, node->left, count );
+  surface = new_string_readable( node->name, NULL );
   decode_hangul( &surface );
-  FOREACH( fs, entry->fs_list ) 
+  FOREACH( feat, lex_node->feat_list ) 
   { 
     (*count)++;
     if (allo_format == NULL) 
     { 
-      value_string = value_to_readable( fs->value, FALSE, 
-					strlen( surface ) + 2 );
+      value_string = value_to_readable( feat->value, FALSE, 
+					g_utf8_strlen( surface, -1 ) + 2 );
       fprintf( stream, "%s: %s\n", surface, value_string );
       free_mem( &value_string );
     } 
     else if (*allo_format != EOS) 
     { 
-      value_string = value_to_readable( fs->value, FALSE, -1 );
+      value_string = value_to_readable( feat->value, FALSE, -1 );
       line_number = int_to_string( *count );
       buffer = replace_arguments( allo_format, "sfn", 
                                   surface, value_string, line_number );
@@ -578,131 +587,133 @@ print_lex_tree_local( FILE *stream, string_t allo_format, lex_node_t *entry,
     }
   }
   free_mem( &surface );
-  print_lex_tree_local( stream, allo_format, 
-                        (lex_node_t *) entry->node.right, count );
+  print_lex_tree( stream, allo_format, node->right, count );
 }
 
 /*---------------------------------------------------------------------------*/
 
 void 
-print_lex_tree( FILE *stream, string_t allo_format )
-/* Print all lexicon entries to STREAM using format ALLO_FORMAT.
+print_lex_buffer( FILE *stream, string_t allo_format )
+/* Print all lexicon entries in the buffer to STREAM using format ALLO_FORMAT.
  * If ALLO_FORMAT == NULL, format the feature structures. */
 { 
   int_t count;
 
   count = 0;
-  print_lex_tree_local( stream, allo_format, lex_tree, &count );
+  print_lex_tree( stream, allo_format, lex_tree, &count );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-count_trie_entries( lex_node_t *tree, int_t *surf_count, int_t *fs_count )
+count_trie_entries( avln_node_t *tree, int_t *surf_count, int_t *feat_count )
 /* Add to *SURF_COUNT the number of valid entries in subtree TREE.
- * Add to *FS_COUNT the number of feature structures in subtree TREE. */
+ * Add to *FEAT_COUNT the number of feature structures in subtree TREE. */
 { 
-  fs_node_t *fs;
-  
+  feat_node_t *feat;
+  lex_node_t *lex_node = (lex_node_t *) tree;
+
   if (tree == NULL) 
     return;
-  count_trie_entries( (lex_node_t *) tree->node.left, surf_count, fs_count );
-  if (tree->fs_list.first != NULL) 
+  count_trie_entries( tree->left, surf_count, feat_count );
+  if (lex_node->feat_list.first != NULL) 
     (*surf_count)++;
-  FOREACH( fs, tree->fs_list ) 
-    (*fs_count)++;
-  count_trie_entries( (lex_node_t *) tree->node.right, surf_count, fs_count );
+  FOREACH( feat, lex_node->feat_list ) 
+    (*feat_count)++;
+  count_trie_entries( tree->right, surf_count, feat_count );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-fill_trie_entries( lex_node_t *tree, trie_entry_t *trie_entries,
-		   int_t *fs_lists, 
-                   int_t *surf_count, int_t *fs_count, int_t *cell_count )
+fill_trie_entries( avln_node_t *tree, trie_entry_t *trie_entries,
+		   int_t *feat_lists, 
+                   int_t *surf_count, int_t *feat_count, int_t *cell_count )
 /* Fill the entries in TREE and its subnodes 
- * into the vectors TRIE_ENTRIES and FS_LISTS.
- * Increment *SURF_COUNT, *FS_COUNT and *CELL_COUNT accordingly. */
+ * into the vectors TRIE_ENTRIES and FEAT_LISTS.
+ * Increment *SURF_COUNT, *FEAT_COUNT and *CELL_COUNT accordingly. */
 { 
-  fs_node_t *fs;
+  feat_node_t *feat;
+  lex_node_t *lex_node = (lex_node_t *) tree;
 
   if (tree == NULL) 
     return;
-  fill_trie_entries( (lex_node_t *) tree->node.left, trie_entries, 
-                     fs_lists, surf_count, fs_count, cell_count );
-  if (tree->fs_list.first != NULL) 
+  fill_trie_entries( tree->left, trie_entries, feat_lists, surf_count, 
+		     feat_count, cell_count );
+  if (lex_node->feat_list.first != NULL) 
   { 
-    trie_entries[ *surf_count ].key = tree->node.name;
-    trie_entries[ *surf_count ].content = *fs_count;
+    trie_entries[ *surf_count ].key = tree->name;
+    trie_entries[ *surf_count ].content = *feat_count;
     (*surf_count)++;
-    FOREACH( fs, tree->fs_list ) 
+    FOREACH( feat, lex_node->feat_list ) 
     { 
-      fs_lists[ *fs_count ] = *cell_count;
-      (*fs_count)++;
-      *cell_count += length_of_value( fs->value );
+      feat_lists[ *feat_count ] = *cell_count;
+      (*feat_count)++;
+      *cell_count += length_of_value( feat->value );
     }
-    MARK_LAST_ENTRY( fs_lists[ *fs_count - 1 ] );
+    MARK_LAST_ENTRY( feat_lists[ *feat_count - 1 ] );
   }
-  fill_trie_entries( (lex_node_t *) tree->node.right, trie_entries, 
-                     fs_lists, surf_count, fs_count, cell_count );
+  fill_trie_entries( tree->right, trie_entries, feat_lists, surf_count, 
+		     feat_count, cell_count );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-write_fs_table( lex_node_t *tree, FILE *stream, string_t file_name )
+write_feat_table( avln_node_t *tree, FILE *stream, string_t file_name )
 { 
-  fs_node_t *fs;
+  feat_node_t *feat;
+  lex_node_t *lex_node = (lex_node_t *) tree;
 
   if (tree == NULL) 
     return;
-  write_fs_table( (lex_node_t *) tree->node.left, stream, file_name );
-  FOREACH( fs, tree->fs_list ) 
+  write_feat_table( tree->left, stream, file_name );
+  FOREACH( feat, lex_node->feat_list ) 
   { 
-    write_vector( fs->value, sizeof( cell_t ), length_of_value( fs->value ),
-		  stream, file_name );
+    write_vector( feat->value, sizeof( cell_t ), 
+		  length_of_value( feat->value ), stream, file_name );
   }
-  write_fs_table( (lex_node_t *) tree->node.right, stream, file_name );
+  write_feat_table( tree->right, stream, file_name );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-write_surfaces( lex_node_t *tree, FILE *stream, string_t file_name )
+write_surfaces( avln_node_t *tree, FILE *stream, string_t file_name )
 { 
   if (tree == NULL) 
     return;
-  write_surfaces( (lex_node_t *) tree->node.left, stream, file_name );
-  write_vector( tree->node.name, sizeof( char_t ), 
-		strlen( tree->node.name ) + 1, stream, file_name );
-  write_surfaces( (lex_node_t *) tree->node.right, stream, file_name );
+  write_surfaces( tree->left, stream, file_name );
+  write_vector( tree->name, sizeof( char_t ), strlen( tree->name ) + 1, 
+		stream, file_name );
+  write_surfaces( tree->right, stream, file_name );
 }
 
 /*---------------------------------------------------------------------------*/
 
 void 
-write_lex_tree( string_t file_name )
-/* Write lexicon tree to file FILE_NAME. */
+write_lex_buffer( string_t file_name )
+/* Write lexicon buffer to file FILE_NAME as a run time lexicon. */
 { 
   lexicon_header_t header;
   FILE *stream;
   pool_t trie_pool;
   int_t trie_root;
-  trie_entry_t *trie_entries; /* Trie entries used to build the trie. */
-  int_t *fs_lists;
-  int_t surf_count, fs_count, cell_count;
+  trie_entry_t *trie_entries; /* Entries used to build the trie. */
+  int_t *feat_lists;
+  int_t surf_count, feat_count, cell_count;
 
   /* Count the number of surfaces and feature structures in the lexicon 
    * and get vectors that will get these values. */
-  surf_count = fs_count = 0;
-  count_trie_entries( lex_tree, &surf_count, &fs_count );
-  fs_lists = new_vector( sizeof( int_t ), fs_count );
+  surf_count = feat_count = 0;
+  count_trie_entries( lex_tree, &surf_count, &feat_count );
+  feat_lists = new_vector( sizeof( int_t ), feat_count );
 
   /* Create the lexicon trie. */
   trie_entries = new_vector( sizeof( trie_entry_t ), surf_count );
-  surf_count = fs_count = cell_count = 0;
-  fill_trie_entries( lex_tree, trie_entries, fs_lists, 
-                     &surf_count, &fs_count, &cell_count );
+  surf_count = feat_count = cell_count = 0;
+  fill_trie_entries( lex_tree, trie_entries, feat_lists, 
+                     &surf_count, &feat_count, &cell_count );
   new_trie( surf_count, trie_entries, &trie_pool, &trie_root );
   free_mem( &trie_entries );
 
@@ -713,74 +724,76 @@ write_lex_tree( string_t file_name )
   set_header( &header.common_header, LEXICON_FILE, LEXICON_CODE_VERSION );
   header.trie_size = pool_item_count( trie_pool );
   header.trie_root = trie_root;
-  header.fs_lists_size = fs_count;
+  header.feat_lists_size = feat_count;
   header.values_size = cell_count;
 
   /* Write everything to the file. */
   write_vector( &header, sizeof( lexicon_header_t ), 1, stream, file_name );
   write_pool( trie_pool, stream, file_name );
-  write_vector( fs_lists, sizeof( int_t ), fs_count, stream, file_name );
-  write_fs_table( lex_tree, stream, file_name );
+  write_vector( feat_lists, sizeof( int_t ), feat_count, stream, file_name );
+  write_feat_table( lex_tree, stream, file_name );
   close_stream( &stream, file_name );
 
   /* Clean up. */
   free_pool( &trie_pool );
-  free_mem( &fs_lists );
+  free_mem( &feat_lists );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-count_prelex_entries( lex_node_t *tree, int_t *entry_count_p, 
+count_prelex_entries( avln_node_t *tree, int_t *entry_count_p, 
 		      int_t *values_size_p, int_t *strings_size_p )
 /* Add to *ENTRY_COUNT_P the number of entries in subtree TREE.
  * Add to *VALUES_SIZE_P the number of cells for feature structures in subtree
  * TREE.
  * Add to *STRINGS_SIZE_P the number of chars for surfaces in subtree TREE. */
 { 
-  fs_node_t *fs;
-  
+  feat_node_t *feat;
+  lex_node_t *lex_node = (lex_node_t *) tree;
+
   if (tree == NULL) 
     return;
-  count_prelex_entries( (lex_node_t *) tree->node.left, 
+  count_prelex_entries( tree->left, 
 			entry_count_p, values_size_p, strings_size_p );
-  if (tree->fs_list.first != NULL) 
-    (*strings_size_p) += strlen( tree->node.name ) + 1;
-  FOREACH( fs, tree->fs_list )
+  if (lex_node->feat_list.first != NULL) 
+    (*strings_size_p) += strlen( tree->name ) + 1;
+  FOREACH( feat, lex_node->feat_list )
   {
     (*entry_count_p)++;
-    (*values_size_p) += length_of_value( fs->value );
+    (*values_size_p) += length_of_value( feat->value );
   }
-  count_prelex_entries( (lex_node_t *) tree->node.right,
+  count_prelex_entries( tree->right,
 			entry_count_p, values_size_p, strings_size_p );
 }
 
 /*---------------------------------------------------------------------------*/
 
 static void 
-write_prelex_entries( lex_node_t *tree, int_t *values_size_p, 
+write_prelex_entries( avln_node_t *tree, int_t *values_size_p, 
 		      int_t *strings_size_p, FILE *stream, string_t file_name )
 /* Write prelex entries in subtree TREE to STREAM.
  * Update *VALUES_SIZE_P and *STRINGS_SIZE_P for indexing.
  * Use FILE_NAME for error messages. */
 { 
-  fs_node_t *fs;
+  feat_node_t *feat;
   prelex_entry_t entry;
+  lex_node_t *lex_node = (lex_node_t *) tree;
 
   if (tree == NULL) 
     return;
-  write_prelex_entries( (lex_node_t *) tree->node.left, values_size_p, 
+  write_prelex_entries( tree->left, values_size_p, 
 			strings_size_p, stream, file_name );
-  FOREACH( fs, tree->fs_list )
+  FOREACH( feat, lex_node->feat_list )
   {
     entry.surface = (*strings_size_p);
-    entry.fs = (*values_size_p);
+    entry.feat = (*values_size_p);
     write_vector( &entry, sizeof( prelex_entry_t ), 1, stream, file_name);
-    (*values_size_p) += length_of_value( fs->value );
+    (*values_size_p) += length_of_value( feat->value );
   }
-  if (tree->fs_list.first != NULL) 
-    (*strings_size_p) += strlen( tree->node.name ) + 1;
-  write_prelex_entries( (lex_node_t *) tree->node.right, values_size_p, 
+  if (lex_node->feat_list.first != NULL) 
+    (*strings_size_p) += strlen( tree->name ) + 1;
+  write_prelex_entries( tree->right, values_size_p, 
 			strings_size_p, stream, file_name );
 }
 
@@ -813,7 +826,7 @@ write_prelex_file( string_t file_name )
   values_size = strings_size = 0;
   write_prelex_entries( lex_tree, &values_size, &strings_size, 
 			stream, file_name );
-  write_fs_table( lex_tree, stream, file_name );
+  write_feat_table( lex_tree, stream, file_name );
   write_surfaces( lex_tree, stream, file_name );
   close_stream( &stream, file_name );
 }
@@ -832,7 +845,7 @@ read_prelex_file( string_t file_name )
   char_t *strings;
   int_t i;
 
-  /* Map stream into main memory. */
+  /* Map file into main memory. */
   map_file( file_name, &prelex_data, &prelex_length );
   header = (prelex_header_t *) prelex_data;
   check_header( &header->common_header, file_name,
@@ -843,7 +856,7 @@ read_prelex_file( string_t file_name )
 
   /* Enter entries into tree. */
   for (i = 0; i < header->entry_count; i++)
-    lex_add_allo( strings + entries[i].surface, values + entries[i].fs );
+    lex_add_allo( strings + entries[i].surface, values + entries[i].feat );
 
   prelex_count = header->entry_count;
   unmap_file( &prelex_data, prelex_length );
@@ -1040,12 +1053,12 @@ read_lex_constants( string_t lexicon_name )
 /*---------------------------------------------------------------------------*/
 
 void 
-generate_allos_for_string( string_t fs_string )
-/* Generate allomorphs from FS_STRING, which should contain a readable feature
- * structure.  Write allomorphs into lexicon buffer. */
+generate_allos_for_string( string_t feat_string )
+/* Generate allomorphs from FEAT_STRING, which should contain a readable
+ * feature structure.  Write allomorphs into lexicon buffer. */
 { 
   free_lex_tree();
-  set_scanner_input( fs_string );
+  set_scanner_input( feat_string );
   TRY 
   { 
     parse_value();
@@ -1058,7 +1071,7 @@ generate_allos_for_string( string_t fs_string )
   END_TRY;
   execute_allo_rule();
   if (! rule_successful) 
-    printf( "Warning: no allomorphs generated.\n" );
+    printf( "Warning: No allomorphs generated.\n" );
   execute_output_filter();
 }
 
@@ -1073,10 +1086,10 @@ init_lex_compiler( string_t allo_rule_file )
 
   /* Initialise lexicon tree. */
   lex_tree = NULL;
-  clear_list( &fs_free_list );
+  clear_list( &feat_free_list );
   lex_node_pool = new_pool( sizeof( lex_node_t ) );
   string_pool = new_pool( sizeof( char_t ) );
-  fs_node_pool = new_pool( sizeof( fs_node_t ) );
+  feat_node_pool = new_pool( sizeof( feat_node_t ) );
   lex_entry_count = allomorph_count = 0;
   prelex_count = intermediate_count = -1;
 }
@@ -1090,9 +1103,9 @@ terminate_lex_compiler( void )
   free_rule_sys( &allo_rule_sys );
   free_lex_tree();
   free_pool( &string_pool );
-  free_pool( &fs_node_pool );
+  free_pool( &feat_node_pool );
   free_pool( &lex_node_pool );
-  clear_list( &fs_free_list );
+  clear_list( &feat_free_list );
   free_const_node( &const_tree );
 }
 
