@@ -1096,6 +1096,8 @@ main( int argc, char *argv[] )
       if (input != NULL) 
 	complain( "Redundant \"-input\"." );
       input = argv[i];
+      if (! g_utf8_validate( input, -1, NULL ))
+	complain( "Illegal UTF-8 character." );
     }
     else if (strcmp_no_case( argv[i], "-quoted" ) == 0
 	     || strcmp_no_case( argv[i], "-q" ) == 0)
