@@ -621,14 +621,14 @@ init_malaga( string_t project_file )
   options = malaga_options;
   execute_set_commands( project_file, "malaga:" );
   malagarc_path = NULL;
-#ifdef UNIX
+#ifdef POSIX
   TRY 
     malagarc_path = absolute_path( "~/.malagarc", NULL );
   IF_ERROR 
     RESUME;
   END_TRY;
 #endif
-#ifdef WINDOWS
+#ifdef WIN32
   TRY 
     malagarc_path = absolute_path( "~\\malaga.ini", NULL );
   IF_ERROR 
