@@ -1,4 +1,4 @@
-/* Copyright (C) 1995 Bjoern Beutel. */
+/* Copyright (C) 1996 Bjoern Beutel. */
 
 /* Description. =============================================================*/
 
@@ -28,6 +28,6 @@ extern command_t auto_variables_option;
 extern command_t switch_option;
 /* Set switches that can be read from Malaga rules. */
 
-extern command_t use_ksc_option; /* Select KSC5601 or Roman output. */
+extern command_t roman_hangul_option; /* Select latin Hangul transcription. */
 
 /* End of file. =============================================================*/
