@@ -1,4 +1,4 @@
-/* Copyright (C) 1995 Bjoern Beutel. */
+/* Copyright (C) 2002 Bjoern Beutel. */
 
 /* Description. =============================================================*/
 
@@ -7,7 +7,7 @@
 /* Variables. ===============================================================*/
 
 extern rectangle_t result_geometry;
-extern string_t result_font_name;
+extern string_t result_font_family;
 extern int_t result_font_size;
 
 /* Functions. ===============================================================*/
