@@ -11,6 +11,7 @@
 #include <string.h>
 #include <time.h>
 #include <setjmp.h>
+#include <glib.h>
 #include "basic.h"
 #include "pools.h"
 #include "values.h"
@@ -56,6 +57,7 @@ typedef struct /* A label for a "foreach" statement. */
 { 
   list_node_t *next;
   string_t name; /* Label of this loop or NULL. */
+  list_t repeat_jumps; /* Pointers to instrs that repeat this loop. */
   list_t out_jumps; /* Pointers to instrs that jump out of this loop. */
 } loop_node_t;
 
@@ -86,7 +88,8 @@ static list_t loop_list; /* List of loops currently open. */
 
 static struct /* The pattern variables defined in a "matches" condition. */
 {
-  int_t count; /* Number of pattern variables */
+  int_t count; /* Number of pattern variables. 
+		* Will be reset when variables are defined. */
   string_t name[ PATTERN_VAR_MAX ]; /* Names of the pattern variables. */
 } pattern_vars;
 
@@ -224,7 +227,8 @@ negate_jump( instr_t *jump )
   case INS_JUMP_IF_NO: 
     opcode = INS_JUMP_IF_YES; 
     break;
-  default: break;
+  default: 
+    break;
   }
   *jump = INSTR( opcode, INSTR_INFO( *jump ) );
 }
@@ -394,18 +398,20 @@ convert_to_condition( expr_type_t *type, condition_t *condition )
 /*---------------------------------------------------------------------------*/
 
 static expr_type_t 
-parse_value( void )
-/* Parse any value. */
+parse_value_no_flush( void )
+/* Parse any value. Convert conditions into yes/no symbols. Do *not* flush the
+ * instruction buffer. */
 {
   expr_type_t type;
   condition_t condition;
+  int_t pattern_var_count;
 
-  pattern_vars.count = 0;
+  pattern_var_count = pattern_vars.count;
   type = parse_value_local( &condition );
+  if (pattern_vars.count > pattern_var_count) 
+    complain( "Variable definition not allowed in values." );
   if (type == CONDITION) 
   { 
-    if (pattern_vars.count != 0) 
-      complain( "No variable definition allowed in values." );
     negate_jump( condition.last_jump );
     patch_instr_list( &condition.true_jumps, code.instr_count );
     emit_instr( INS_PUSH_SYMBOL, YES_SYMBOL );
@@ -417,10 +423,21 @@ parse_value( void )
     return VALUE;
   } 
   else 
-  { 
-    flush_buffer();
     return type;
-  }
+}
+
+/*---------------------------------------------------------------------------*/
+
+static expr_type_t
+parse_value( void )
+/* Parse any value. Convert conditions into yes/no symbols. Flush the
+ * instruction buffer. */
+{
+  expr_type_t type;
+
+  type = parse_value_no_flush();
+  flush_buffer();
+  return type;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -432,7 +449,6 @@ parse_constant_value( int_t *value_index )
   expr_type_t type;
   condition_t condition;
 
-  pattern_vars.count = 0;
   type = parse_value_local( &condition );
   if (type != CONST_VALUE) 
     complain( "Constant value expected." );
@@ -447,7 +463,6 @@ parse_condition( condition_t *condition )
 {
   expr_type_t type;
   
-  pattern_vars.count = 0;
   type = parse_expression( condition );
   convert_to_condition( &type, condition );
 }
@@ -519,6 +534,56 @@ parse_subrule_call( string_t ident )
 
 /*---------------------------------------------------------------------------*/
 
+static void 
+parse_if_expression( void )
+/* Parse an "if" expression. */
+{
+  condition_t condition;
+  list_t end_jumps;
+  list_t else_jumps;
+  instr_t *end_jump;
+  int_t pattern_var_count;
+
+  clear_list( &else_jumps );
+  clear_list( &end_jumps );
+  test_token( TOK_IF );
+  do 
+  { 
+    read_next_token();
+
+    /* Jump here if previous conditions were false. */
+    patch_instr_list( &else_jumps, code.instr_count );
+    pattern_var_count = pattern_vars.count;
+    parse_condition( &condition );
+    if (pattern_vars.count > pattern_var_count)
+      complain( "No variable definitions allowed in if-expressions." );
+    parse_token( TOK_THEN );
+
+    /* Fall through if condition is true. */
+    else_jumps = condition.false_jumps;
+    negate_jump( condition.last_jump );
+    add_instr_to_list( &else_jumps, condition.last_jump );
+    patch_instr_list( &condition.true_jumps, code.instr_count );
+    parse_value();
+    code.stack_index--;
+
+    /* Emit jump to end. */
+    end_jump = emit_instr( INS_JUMP, 0 );
+    add_instr_to_list( &end_jumps, end_jump );
+  } while (next_token == TOK_ELSEIF);
+
+  parse_token( TOK_ELSE );
+  patch_instr_list( &else_jumps, code.instr_count );
+  parse_value();
+
+  parse_token( TOK_END );
+  if (next_token == TOK_IF) 
+    read_next_token();
+  patch_instr_list( &end_jumps, code.instr_count );
+}
+
+/*---------------------------------------------------------------------------*/
+
 static expr_type_t 
 parse_simple_value( condition_t *condition )
 /* Parse a simple value or a condition in parentheses.
@@ -541,10 +606,8 @@ parse_simple_value( condition_t *condition )
     { 
       while (TRUE) 
       { 
-	elem_type = parse_value_local( condition );
-        if (elem_type == CONDITION) 
-	  complain( "No condition allowed in lists." );
-	else if (elem_type == VALUE) 
+	elem_type = parse_value_no_flush();
+	if (elem_type == VALUE) 
 	  type = VALUE;
         i++;
         if (next_token != ',') 
@@ -566,14 +629,12 @@ parse_simple_value( condition_t *condition )
       { 
 	elem_type = parse_value_local( condition );
         if (elem_type == CONDITION) 
-	  complain( "No condition allowed in records." );
+	  complain( "No conditions allowed as attribute names." );
 	else if (elem_type == VALUE) 
 	  type = VALUE;
         parse_token( ':' );
-        elem_type = parse_value_local( condition );
-        if (elem_type == CONDITION) 
-	  complain( "No condition allowed in records." );
-	else if (elem_type == VALUE) 
+        elem_type = parse_value_no_flush();
+	if (elem_type == VALUE) 
 	  type = VALUE;
         i++;
         if (next_token != ',') 
@@ -624,6 +685,9 @@ parse_simple_value( condition_t *condition )
     type = parse_expression( condition );
     parse_token( ')' );
     return type;
+  case TOK_IF:
+    parse_if_expression();
+    return VALUE;
   default:
     complain( "Value expected, not %s.", token_as_text( next_token ) );
   }
@@ -642,7 +706,7 @@ parse_unary_value( condition_t *condition )
     read_next_token();
     type = parse_simple_value( condition );
     if (type == CONDITION) 
-      complain( "No condition allowed after unary \"-\"." );
+      complain( "Conditions not allowed after unary \"-\"." );
     buffer_instr( INS_UNARY_MINUS_OP, 0 );
   } 
   else 
@@ -667,7 +731,7 @@ parse_dotted_value( condition_t *condition )
     read_next_token();
     elem_type = parse_unary_value( condition );
     if (elem_type == CONDITION) 
-      complain( "No condition allowed after \".\"." );
+      complain( "Conditions not allowed after \".\"." );
     else if (elem_type == VALUE) 
       type = VALUE;
     if (elem_type == CONST_VALUE
@@ -700,7 +764,7 @@ parse_term_value( condition_t *condition )
     read_next_token();
     elem_type = parse_dotted_value( condition );
     if (elem_type == CONDITION) 
-      complain( "No condition allowed after \"*\" and \"/\"." );
+      complain( "Conditions not allowed after \"*\" and \"/\"." );
     else if (elem_type == VALUE) 
       type = VALUE;
     if (operator_token == '*') 
@@ -733,7 +797,7 @@ parse_value_local( condition_t *condition )
     { 
       elem_type = parse_term_value( condition );
       if (elem_type == CONDITION) 
-	complain( "No condition allowed after \"-\"." );
+	complain( "Conditions not allowed after \"-\"." );
       else if (elem_type == VALUE) 
 	type = VALUE;
       if (elem_type == CONST_VALUE
@@ -749,7 +813,7 @@ parse_value_local( condition_t *condition )
     { 
       elem_type = parse_term_value( condition );
       if (elem_type == CONDITION) 
-	complain( "No condition allowed after \"+\"." );
+	complain( "Conditions not allowed after \"+\"." );
       else if (elem_type == VALUE) 
 	type = VALUE;
       buffer_instr( INS_PLUS_OPERATION, 0 );
@@ -835,6 +899,7 @@ define_pattern_vars( void )
     define_var_in_scope( pattern_vars.name[i], code.stack_index - 1 );
     free_mem( &pattern_vars.name[i] );
   }
+  pattern_vars.count = 0;
 }
 
 /*---------------------------------------------------------------------------*/
@@ -903,7 +968,7 @@ parse_comparison( condition_t *condition )
   read_next_token();
   type2 = parse_value_local( condition );
   if (type2 == CONDITION) 
-    complain( "No condition allowed in comparisons." );
+    complain( "Conditions not allowed in comparisons." );
 
   /* Emit code and produce CONDITION. */
   clear_list( &condition->true_jumps );
@@ -927,7 +992,7 @@ parse_simple_condition( condition_t *condition )
   { 
     read_next_token();
     type = parse_comparison( &condition2 );
-    if (pattern_vars.count != pattern_var_count) 
+    if (pattern_vars.count > pattern_var_count) 
       complain( "No variable definition allowed in negations." );
     convert_to_condition( &type, &condition2 );
     negate_jump( condition2.last_jump );
@@ -972,7 +1037,7 @@ parse_expression( condition_t *condition )
       add_instr_to_list( &condition->true_jumps, condition->last_jump );
       condition->last_jump = condition2.last_jump;
     }
-    if (pattern_vars.count != pattern_var_count) 
+    if (pattern_vars.count > pattern_var_count) 
       complain( "No variable definition allowed in disjunctions." );
   } 
   else if (next_token == TOK_AND) 
@@ -1015,7 +1080,7 @@ parse_attribute_path( void )
   { 
     type = parse_unary_value( &condition );
     if (type == CONDITION) 
-      complain( "No condition allowed in paths." );
+      complain( "Conditions not allowed in paths." );
     path_length++;
     if (next_token != '.') 
       break;
@@ -1196,6 +1261,33 @@ parse_choose_statement( void )
 
 /*---------------------------------------------------------------------------*/
 
+static void 
+parse_continue_statement( void )
+/* Parse a "continue" statement. */
+{
+  loop_node_t *loop;
+
+  new_source_line();
+  parse_token( TOK_CONTINUE );
+  if (next_token == TOK_IDENT) 
+  { 
+    FOREACH( loop, loop_list ) 
+    { 
+      if (loop->name != NULL && strcmp_no_case( token_name, loop->name ) == 0)
+	break;
+    }
+    read_next_token();
+  } 
+  else 
+    loop = (loop_node_t *) loop_list.first;
+  if (loop == NULL) 
+    complain( "\"continue\" must be within a loop." );
+  add_instr_to_list( &loop->repeat_jumps, emit_instr( INS_JUMP, 0 ) );
+  parse_token( ';' );
+}
+
+/*---------------------------------------------------------------------------*/
+
 static void 
 parse_define_statement( void )
 /* Parse a "define" statement. */
@@ -1255,8 +1347,8 @@ parse_foreach_statement( void )
 {
   string_t loop_name; /* Name of loop. */
   string_t var_name; /* Name of iteration variable. */
-  int_t list_index; /* Stack index of list variable. */
-  int_t label; /* Label where to jump back. */
+  int_t label; /* PC to jump to when repeating loop. */
+  int_t list_index;
   loop_node_t *loop;
 
   new_source_line();
@@ -1277,6 +1369,7 @@ parse_foreach_statement( void )
   /* Create loop list. */
   loop = new_node( &loop_list, sizeof( loop_node_t ), LIST_START );
   loop->name = loop_name;
+  clear_list( &loop->repeat_jumps );
   clear_list( &loop->out_jumps );
   parse_token( TOK_FOREACH );
   parse_var_name( &var_name );
@@ -1305,6 +1398,13 @@ parse_foreach_statement( void )
   parse_token( ';' );
   no_source_line();
 
+  /* Patch the jumps to repeat the loop. */
+  if (loop->repeat_jumps.first != NULL)
+  { 
+    patch_instr_list( &loop->repeat_jumps, code.instr_count );
+    emit_instr( INS_POP_TO, list_index + 2 );
+  }
+
   /* Iterate variables. */
   emit_instr( INS_ITERATE, list_index + 1 );
   emit_instr( INS_JUMP, label );
@@ -1314,7 +1414,9 @@ parse_foreach_statement( void )
   free_first_node( &loop_list );
   free_mem( &loop_name );
 
-  emit_instr( INS_POP_TO, list_index ); /* Pop iteration variable and list. */
+  /* Pop iteration variable and list. */
+  emit_instr( INS_POP_TO, list_index ); 
+
   free_mem( &var_name );
 }
 
@@ -1455,18 +1557,21 @@ parse_list_assignment( void )
 /*---------------------------------------------------------------------------*/
 
 static void 
-parse_parallel_statement( void )
-/* Parse a "parallel" statement. */
+parse_select_statement( void )
+/* Parse a "select" or "parallel" statement. */
 {
   list_t end_jumps;
   instr_t *last_jump;
+  int_t start_token;
 
   last_jump = NULL;
   clear_list( &end_jumps );
-  test_token( TOK_PARALLEL );
-  do 
+  if (next_token != TOK_SELECT && next_token != TOK_PARALLEL)
+    complain( "Missing \"select\" or \"parallel\"." );
+  start_token = next_token;
+  do
   { 
-    read_next_token(); /* Read over "parallel" or "and". */
+    read_next_token(); /* Read over "select, "parallel", "or" or "and". */
 
     if (last_jump != NULL) 
     { 
@@ -1481,14 +1586,15 @@ parse_parallel_statement( void )
     open_scope();
     parse_statements();
     close_scope( TRUE );
-  } while (next_token == TOK_AND);
+  } while ((start_token == TOK_PARALLEL && next_token == TOK_AND)
+	   || (start_token == TOK_SELECT && next_token == TOK_OR));
 
   /* The last patch is invalid, replace it by an INS_NOP. */
   *last_jump = INSTR( INS_NOP, 0 );
 
   patch_instr_list( &end_jumps, code.instr_count );
   parse_token( TOK_END );
-  if (next_token == TOK_PARALLEL) 
+  if (next_token == start_token) 
     read_next_token();
   parse_token( ';' );
 }
@@ -1663,6 +1769,9 @@ parse_statements( void )
     case TOK_CHOOSE: 
       parse_choose_statement(); 
       break;
+    case TOK_CONTINUE:
+      parse_continue_statement();
+      break;
     case TOK_DEFINE: 
       parse_define_statement(); 
       break;
@@ -1683,8 +1792,9 @@ parse_statements( void )
     case '<': 
       parse_list_assignment(); 
       break;
+    case TOK_SELECT:
     case TOK_PARALLEL: 
-      parse_parallel_statement(); 
+      parse_select_statement(); 
       break;
     case TOK_REPEAT: 
       parse_repeat_statement(); 
@@ -1752,8 +1862,8 @@ parse_rule( void )
     rule_type = FILTER_RULE;
     break;
   case TOK_PRUNING_RULE:
-    if (code.file_type != SYNTAX_RULE_FILE) 
-      complain( "\"pruning_rule\" only allowed in syntax rule files." );
+    if (code.file_type == ALLO_RULE_FILE) 
+      complain( "\"pruning_rule\" not allowed in allo rule files." );
     rule_type = PRUNING_RULE;
     break;
   case TOK_ROBUST_RULE:
@@ -1812,35 +1922,33 @@ parse_rule( void )
 				param_count, rule_line, rule_file );
 
   /* Save rule number if this is a special rule. */
-  if (rule_token == TOK_PRUNING_RULE) 
-  { 
+  switch (rule_token)
+  {
+  case TOK_PRUNING_RULE:
     if (code.pruning_rule != -1) 
       complain( "\"pruning_rule\" defined twice." );
     code.pruning_rule = code.rule_count;
-  } 
-  else if (rule_token == TOK_ROBUST_RULE) 
-  { 
+    break;
+  case TOK_ROBUST_RULE:
     if (code.robust_rule != -1) 
       complain( "\"robust_rule\" defined twice." );
     code.robust_rule = code.rule_count;
-  }
-  else if (rule_token == TOK_ALLO_RULE) 
-  { 
+    break;
+  case TOK_ALLO_RULE:
     if (code.allo_rule != -1) 
       complain( "\"allo_rule\" defined twice." );
     code.allo_rule = code.rule_count;
-  } 
-  else if (rule_token == TOK_INPUT_FILTER) 
-  { 
+    break;
+  case TOK_INPUT_FILTER:
     if (code.input_filter != -1) 
       complain( "\"input_filter\" defined twice." );
     code.input_filter = code.rule_count;
-  }
-  else if (rule_token == TOK_OUTPUT_FILTER) 
-  { 
+   break;
+  case TOK_OUTPUT_FILTER:
     if (code.output_filter != -1) 
       complain( "\"output_filter\" defined twice." );
     code.output_filter = code.rule_count;
+    break;
   }
   parse_token( ':' );
 
@@ -1937,7 +2045,7 @@ parse_rules( void )
       if (code.initial_rule_set != -1) 
 	complain( "Initial state is already defined." );
       read_next_token();
-      parse_constant_value( &code.initial_fs );
+      parse_constant_value( &code.initial_feat );
       parse_token( ',' );
       parse_rule_set( &code.initial_rule_set );
       parse_token( ';' );
