@@ -6,11 +6,11 @@
 
 /* Includes. ================================================================*/
 
-#include <ctype.h>
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
 #include <setjmp.h>
+#include <glib.h>
 #include "basic.h"
 #include "files.h"
 #include "scanner.h"
@@ -28,6 +28,7 @@ static struct { string_t name; int_t code; } keywords[ NUMBER_OF_KEYWORDS ] =
   { "break", TOK_BREAK },
   { "choose", TOK_CHOOSE },
   { "combi_rule", TOK_COMBI_RULE },
+  { "continue", TOK_CONTINUE },
   { "default", TOK_DEFAULT },
   { "define", TOK_DEFINE },
   { "else", TOK_ELSE },
@@ -57,6 +58,7 @@ static struct { string_t name; int_t code; } keywords[ NUMBER_OF_KEYWORDS ] =
   { "return", TOK_RETURN },
   { "robust_rule", TOK_ROBUST_RULE },
   { "rules", TOK_RULES },
+  { "select", TOK_SELECT },
   { "stop", TOK_STOP },
   { "subrule", TOK_SUBRULE },
   { "then", TOK_THEN },
@@ -70,6 +72,8 @@ typedef struct /* A source stream for lexical analysis. */
   list_node_t *next; /* The next (including) source stream. */
   FILE *stream; /* The input stream for this include level. */
   string_t file_name; /* The name of the input file. */
+  text_t *line; /* The current line. */
+  string_t next_char_p; /* Pointer to the next char in LINE to be read. */
   int_t column; /* Column that has been read. */
   int_t line_number; /* Number of the line that has been read. */
   int_t next_char; /* Buffer NEXT_CHAR if this source is backed up. */
@@ -80,7 +84,7 @@ typedef struct /* A source stream for lexical analysis. */
 
 int_t next_token;
 string_t token_name;
-string_t token_string;
+char_t *token_string;
 double token_number;
 
 /* Variables. ===============================================================*/
@@ -90,7 +94,7 @@ static list_t sources; /* The list of sources, current source first. */
 static string_t scanner_input;
 /* If no file is included, the scanner reads its input from SCANNER_INPUT. */
 
-static int_t next_char; /* The next char to be read. */
+static int_t next_char; /* The next unicode char to be read. */
 
 static text_t *token_text; /* The text of the next token. */
 
@@ -104,32 +108,70 @@ read_next_char( void )
  * If reading from stream, update column information. */
 { 
   source_t *source;
+  int_t c;
 
   source = (source_t *) sources.first;
-  if (scanner_input != NULL) 
+  if (scanner_input != NULL) /* Read from a string. */
   { 
-    if (*scanner_input != EOS) 
-      next_char = *scanner_input++;
-    else 
+    if (*scanner_input == EOS)
       next_char = EOF;
+    else
+    {
+      next_char = g_utf8_get_char( scanner_input );
+      scanner_input = g_utf8_next_char( scanner_input );
+    }
   } 
-  else if (source != NULL) 
+  else if (source != NULL) /* Read from a file. */
   { 
-    next_char = getc( source->stream );
-    if (next_char == EOF && ferror( source->stream )) 
-    { 
-      complain( "Can't read from \"%s\": %s.", 
-		source->file_name, strerror( errno ) );
+    /* Read a new line if current line is empty. */
+    if (*source->next_char_p == EOS)
+    {
+      clear_text( source->line );
+      do
+      {
+	c = getc( source->stream );
+	if (c == EOS)
+	  complain( "Null byte in \"%s\"", source->file_name );
+	else if (c == EOF) 
+	{
+	  if (ferror( source->stream ))
+	  {
+	    complain( "Can't read from \"%s\": %s.", 
+		      source->file_name, strerror( errno ) );
+	  }
+	  else 
+	    break;
+	}
+	else
+	  ADD_CHAR_TO_TEXT( source->line, c );
+      } while (c != '\n');
+      
+      if (! g_utf8_validate( source->line->buffer, -1, NULL ))
+	complain( "Illegal UTF-8 character in \"%s\".", source->file_name );
+      source->next_char_p = source->line->buffer;
+    }
+
+    if (*source->next_char_p == EOS)
+      next_char = EOF;
+    else
+    {
+      /* Get next char from current line. */
+      next_char = g_utf8_get_char( source->next_char_p );
+      source->next_char_p = g_utf8_next_char( source->next_char_p );
+      
+      /* Update line and column information. */
+      if (next_char == '\t') 
+	source->column = (source->column + 8) & ~7;
+      else if (next_char == '\n') 
+      { 
+	source->column = 0;
+	source->line_number++;
+      }
+      else if (next_char == '\r')
+	source->column = 0;
+      else if (next_char != EOF) 
+	source->column++;
     }
-    if (next_char == '\t') 
-      source->column = (source->column + 8) & ~7;
-    else if (next_char == '\n') 
-    { 
-      source->column = 0;
-      source->line_number++;
-    } 
-    else if (next_char != EOF) 
-      source->column++;
   } 
   else 
     next_char = EOF;
@@ -227,6 +269,8 @@ begin_include( string_t file_name )
   }
   /* Create new source description. */
   source = new_node( &sources, sizeof( source_t ), LIST_START );
+  source->line = new_text();
+  source->next_char_p = source->line->buffer;
   source->file_name = file_name;
   source->line_number = 1;
   source->column = 0;
@@ -245,6 +289,7 @@ end_include( void )
 
   source = (source_t *) sources.first;
   close_stream( &source->stream, source->file_name );
+  free_text( &source->line );
   free_first_node( &sources );
   if (sources.first != NULL) 
   { 
@@ -283,7 +328,10 @@ terminate_scanner( void )
 
   scanner_input = NULL;
   FOREACH_FREE( source, sources ) 
+  {
     close_stream( &source->stream, NULL );
+    free_text( &source->line );
+  }
   token_name = NULL;
   free_text( &token_text );
   free_mem( &token_string );
@@ -298,11 +346,11 @@ read_name( void )
   token_name = NULL;
   clear_text( token_text );
 
-  while (next_char != EOF &&
-         (next_char == '_' || next_char == '&' || next_char == '|' 
-          || IS_ALPHA( next_char ) || IS_DIGIT( next_char ))) 
+  while (next_char != EOF
+         && (g_unichar_isalnum( next_char )
+	     || next_char == '_' || next_char == '&' || next_char == '|'))
   { 
-    add_char_to_text( token_text, next_char );
+    add_unichar_to_text( token_text, next_char );
     read_next_char();
   }
 
@@ -314,7 +362,7 @@ read_name( void )
 /*---------------------------------------------------------------------------*/
 
 static int_t 
-keyword( string_t name )
+keyword_code( string_t name )
 /* Look up NAME in the keyword table and return its token value.
  * If NAME is no keyword, return TOK_IDENT. */
 {
@@ -347,16 +395,16 @@ read_number( void )
   token_name = NULL;
   clear_text( token_text );
 
-  while (IS_DIGIT( next_char )) 
+  while (next_char >= '0' && next_char <= '9') 
   { 
     add_char_to_text( token_text, next_char );
     read_next_char();
   }
-  if (TO_LOWER( next_char ) == 'l') 
+  if (next_char == 'l' || next_char == 'L') 
     read_next_char();
-  else if (TO_LOWER( next_char ) == 'r') 
+  else if (next_char == 'r' || next_char == 'R') 
   { 
-    insert_in_text( token_text, "-", 0 );
+    insert_char_in_text( token_text, '-', 0 );
     read_next_char();
   } 
   else 
@@ -365,9 +413,9 @@ read_number( void )
     { 
       add_char_to_text( token_text, next_char );
       read_next_char();
-      if (! IS_DIGIT( next_char )) 
+      if (next_char < '0' || next_char >'9') 
 	complain( "Missing digits after \".\"." );
-      while (IS_DIGIT( next_char )) 
+      while (next_char >= '0' && next_char <= '9') 
       { 
 	add_char_to_text( token_text, next_char );
         read_next_char();
@@ -382,9 +430,9 @@ read_number( void )
 	add_char_to_text( token_text, next_char );
         read_next_char();
       }
-      if (! IS_DIGIT( next_char )) 
+      if (next_char < '0' || next_char > '9') 
 	complain( "Missing exponent." );
-      while (IS_DIGIT( next_char )) 
+      while (next_char >= '0' && next_char <= '9') 
       { 
 	add_char_to_text( token_text, next_char );
         read_next_char();
@@ -397,13 +445,64 @@ read_number( void )
 
 /*---------------------------------------------------------------------------*/
 
+static void 
+read_string( void )
+/* Read a string. Save its value in TOKEN_STRING. */
+{
+  int_t i;
+  u_int_t code;
+
+  token_name = NULL;
+  clear_text( token_text );
+  read_next_char(); /* Overread beginning '"'. */
+  while (next_char != '\"') 
+  { 
+    if (next_char == EOF || next_char == '\n') 
+      complain( "Unterminated string at end of line." );
+    if (next_char != '\\') 
+    {
+      add_unichar_to_text( token_text, next_char );
+      read_next_char();
+    }
+    else
+    { 
+      read_next_char();
+      if (next_char == '\\' || next_char == '\"')
+      {
+	add_char_to_text( token_text, next_char );
+	read_next_char();
+      }
+      else if (next_char >= '0'  && next_char <= '7')
+      { 
+	code = 0;
+	for (i = 0; i < 3; i++) 
+	{ 
+	  if (next_char >= '0' && next_char <= '7')
+	    code = 8 * code + (next_char - '0');
+	  else 
+	    complain( "Escape sequence must have 3 octal digits." );
+	  read_next_char();
+	}
+	if (! g_unichar_validate( code ))
+	  complain( "Escape sequence defines invalid character." );
+	add_unichar_to_text( token_text, code );
+      }
+      else 
+	complain( "Illegal escape sequence in string." );
+    }
+  }
+  read_next_char(); /* Read over final '"'. */
+  free_mem( &token_string ); /* Free old token string. */
+  token_string = new_string( token_text->buffer, NULL );
+}
+
+/*---------------------------------------------------------------------------*/
+
 void 
 read_next_token( void )
 /* Read the next token from current source into NEXT_TOKEN.
  * If end of input stream is reached, return EOF. */
 {
-  int_t i, code;
-
   /* Read chars until a token has been recognised. */
   while (TRUE) 
   { 
@@ -417,6 +516,12 @@ read_next_token( void )
     case '\n': /* Read over whitespace. */
       read_next_char();
       break;
+    case '\r':
+      read_next_char();
+      if (next_char != '\n')
+	complain( "Carriage return without line feed." );
+      read_next_char();
+      break;
     case '#': /* Read over a comment. */
       do 
       { 
@@ -424,46 +529,7 @@ read_next_token( void )
       } while (next_char != '\n' && next_char != EOF);
       break;
     case '\"': /* Read a string. */
-      token_name = NULL;
-      clear_text( token_text );
-      read_next_char(); /* Overread beginning '"'. */
-      while (next_char != '\"') 
-      { 
-	if (next_char == EOF || next_char == '\n') 
-	  complain( "Unterminated string at end of line." );
-        if (next_char != '\\') 
-	  add_char_to_text( token_text, next_char );
-	else
-	{ 
-	  read_next_char();
-          if (next_char == '\\' || next_char == '\"')
-            add_char_to_text( token_text, next_char );
-          else if (TO_LOWER( next_char ) == 'x') 
-	  { 
-	    code = 0;
-            for (i = 0; i < 2 * (int_t) sizeof( char_t ); i++) 
-	    { 
-              read_next_char();
-	      if (IS_DIGIT( next_char ))
-		code = 16 * code + ORD( next_char ) - '0';
-	      else if (TO_LOWER( next_char ) >= 'a' 
-		       && TO_LOWER( next_char ) <= 'f')
-	      {
-		code = 16 * code + 10 + TO_LOWER( next_char ) - 'a';
-	      }
-	      else 
-		complain( "Control char sequence must have 2 digits." );
-            }
-            add_char_to_text( token_text, code );
-          } 
-	  else 
-	    complain( "Illegal escape sequence." );
-        } 
-	read_next_char();
-      }
-      read_next_char(); /* Read over final '"'. */
-      free_mem( &token_string ); /* Free old token string. */
-      token_string = new_string( token_text->buffer, NULL );
+      read_string();
       next_token = TOK_STRING;
       return;
     case ':': /* Read a ":", ":=", ":=+", ":=-", ":=*", ":=/". */
@@ -506,7 +572,7 @@ read_next_token( void )
       } 
       else if (next_char == '~') 
       { 
-	next_token = TOK_NOT_CONGRUENT;
+        next_token = TOK_NOT_CONGRUENT;
         read_next_char();
       } 
       else 
@@ -529,11 +595,11 @@ read_next_token( void )
       next_token = TOK_CONSTANT;
       return;
     default: 
-      if (IS_ALPHA( next_char ) 
-          || next_char == '_' || next_char == '&' || next_char == '|') 
+    if (g_unichar_isalpha( next_char ) 
+        || next_char == '_' || next_char == '&' || next_char == '|') 
       { 
 	read_name();
-        next_token = keyword( token_name );
+        next_token = keyword_code( token_name );
         return;
       } 
       else 
