@@ -95,13 +95,13 @@ terminate_transmit( void )
 /*---------------------------------------------------------------------------*/
 
 static void 
-do_transmit_line_option( string_t arguments )
+do_transmit_cmd_option( string_t arguments )
 /* Set the command line to start the transmit process. */
 {
   if (*arguments == EOS) 
   { 
-    printf( "transmit-line: \"%s\"\n", (transmit_process.command_line == NULL ?
-					"" : transmit_process.command_line) );
+    printf( "transmit-cmd: \"%s\"\n", (transmit_process.command_line == NULL ?
+				       "" : transmit_process.command_line) );
   } 
   else 
   { 
@@ -112,11 +112,11 @@ do_transmit_line_option( string_t arguments )
   parse_end( &arguments );
 }
 
-command_t transmit_line_option = 
+command_t transmit_cmd_option = 
 { 
-  "transmit-line transmit", do_transmit_line_option,
-  "Usage: set transmit-line \"TRANSMIT_COMMAND_LINE\"\n"
-  "Set the command line that is used to start the transmit process.\n"
+  "transmit-cmd transmit-line transmit", do_transmit_cmd_option,
+  "Usage: set transmit-cmd \"TRANSMIT_COMMAND_LINE\"\n"
+  "Set the command that is used to start the transmit process.\n"
 };
 
 /*---------------------------------------------------------------------------*/
