@@ -14,7 +14,7 @@ extern void terminate_transmit( void );
 
 /* Commands. ================================================================*/
 
-extern command_t transmit_line_option;
+extern command_t transmit_cmd_option;
 /* Set the command line to start the display process. */
 
 extern command_t transmit_command;
