@@ -286,6 +286,7 @@ expose_path( canvas_t *canvas, rectangle_t *area )
   int_t space_width = get_space_width( canvas );
   int_t font_height = get_font_height( canvas );
   int_t border_width = get_border_width( canvas );
+  int_t line_width = get_line_width( canvas );
 
   set_color( BLACK );
 
@@ -311,8 +312,10 @@ expose_path( canvas_t *canvas, rectangle_t *area )
       {
 	x1 = node->rule_name->x;
 	x2 = x1 + node->rule_name->width + space_width;
-	y = node->rule_name->y + font_height + 1;
+	y = node->rule_name->y + font_height + line_width;
 	draw_pos_string( node->rule_name, canvas );
+
+	/* Draw arrow. */
 	draw_lines( 2, x1, y, x2, y );
 	draw_lines( 3, 
 		    x2 - space_width, y - space_width / 2,
@@ -490,6 +493,7 @@ configure_node( canvas_t *canvas, tree_node_t *node, int_t x, int_t y )
   bool_t is_first;
   int_t font_height = get_font_height( canvas ); 
   int_t space_width = get_space_width( canvas );
+  int_t line_width = get_line_width( canvas );
 
   if (tree_show_state_indexes && node->state_index != -1)
   {
@@ -527,9 +531,9 @@ configure_node( canvas_t *canvas, tree_node_t *node, int_t x, int_t y )
     w = MAX( subnode->pos_link_surf->width, subnode->pos_rule_name->width );
     w = MAX( w, space_width );
     subnode->pos_link_surf->x = x + (w - subnode->pos_link_surf->width) / 2;
-    subnode->pos_link_surf->y = tree_height - font_height - 1;
+    subnode->pos_link_surf->y = tree_height - font_height - line_width;
     subnode->pos_rule_name->x = x + (w - subnode->pos_rule_name->width) / 2;
-    subnode->pos_rule_name->y = tree_height + 1;
+    subnode->pos_rule_name->y = tree_height + line_width;
     configure_node( canvas, subnode, x + w, tree_height );
   }
 }
@@ -568,14 +572,14 @@ static void
 expose_node( canvas_t *canvas, tree_node_t *node, int_t font_height )
 {
   tree_node_t *subnode;
-  int_t x = node->x;
-  int_t y = node->y;
-  int_t radius1 = font_height / 2;
-  int_t radius2 = (radius1 * 3) / 5;
-  int_t old_y;
+  const int_t x = node->x;
+  const int_t y = node->y;
+  const int_t radius1 = font_height / 2;
+  const int_t radius2 = (radius1 * 3) / 5;
+  int_t middle = y;
+  int_t bottom = y;
 
   /* Draw all subnodes. */
-  old_y = y;
   for (subnode = node->child; subnode != NULL; subnode = subnode->sibling)
   {
     if (tree_mode == NO_DEAD_ENDS && subnode->type == BREAK_NODE) 
@@ -583,15 +587,17 @@ expose_node( canvas_t *canvas, tree_node_t *node, int_t font_height )
     if (tree_mode == RESULT_PATHS && ! subnode->result_path) 
       continue;
     
+    /* Draw horizontal line. */
+    bottom = subnode->y;
     if (node->in_path && subnode->in_path)
     {
+      middle = subnode->y;
       set_color( RED );
-      old_y = y;
     }
     else 
       set_color( BLACK );
-    draw_lines( 3, x, old_y, x, subnode->y, subnode->x, subnode->y );
-    old_y = subnode->y + 1;
+    draw_lines( 2, x, subnode->y, subnode->x, subnode->y );
+
     if (path_begin == NULL && subnode->in_path) 
       set_color( RED );
     draw_pos_string( subnode->pos_link_surf, canvas );
@@ -600,6 +606,15 @@ expose_node( canvas_t *canvas, tree_node_t *node, int_t font_height )
     expose_node( canvas, subnode, font_height );
   }
 
+  /* Draw vertical line. */
+  set_color( BLACK );
+  draw_lines( 2, x, middle, x, bottom ); 
+  if (middle != y)
+  {
+    set_color( RED );
+    draw_lines( 2, x, y, x, middle );
+  }
+
   /* Draw the node itself. */
   if (node->type == BREAK_NODE)
   {
