@@ -1,12 +1,19 @@
-/* Copyright (C) 1995 Bjoern Beutel. */
+/* Copyright (C) 2002 Bjoern Beutel. */
 
 /* Description. =============================================================*/
 
 /* Read in and display a Malaga Tree. */
 
+/* Types. ===================================================================*/
+
+typedef enum {FULL_TREE, NO_DEAD_ENDS, RESULT_PATHS} tree_mode_t;
+
 /* Variables. ===============================================================*/
 
 extern rectangle_t tree_geometry, path_geometry;
+extern tree_mode_t tree_mode;
+extern bool_t show_state_indexes;
+extern bool_t inline_path; /* If TRUE, feature structures are shown in line. */
 
 /* Functions. ===============================================================*/
 
